"""CLI for R2X."""

import json
import sys
import pathlib
from .cli_functions import base_cli
from .config import Configuration
from .runner import scenario_runner
from .utils import read_yaml
from .logger import setup_logging


def cli():
    """Entry point for R2X."""
    parser = base_cli()
    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(0)

    args, remaining_args = parser.parse_known_args()

    if "--help" in remaining_args:
        parser.print_help(sys.stderr)
        sys.exit(0)

    args = parser.parse_args()

    setup_logging(debug=args.debug)

    cli_args = {k: v for k, v in vars(args).items() if v is not None}
    user_dict = cli_args.pop("user_dict", None)

    # Try to read file first
    if user_dict is not None:
        if pathlib.Path(user_dict).exists():
            if o_dict := read_yaml(user_dict):
                user_dict = o_dict
        else:
            user_dict = json.loads(user_dict)

    if cli_args.get("cases_file") is None:
        if solve_year := cli_args.get("solve_year"):
            cli_args["solve_year"] = solve_year[0] if len(solve_year) == 1 else solve_year
        config_mgr = Configuration.from_cli(cli_args, user_dict=user_dict)
    else:
        config_mgr = Configuration.from_cases(
            cases_fpath=cli_args["cases_file"], cli_args=cli_args, user_dict=user_dict
        )

    if scenario_name := cli_args.get("scenario_name", None):
        config_mgr = config_mgr.get(scenario_name)

    scenario_runner(config_mgr)


if __name__ == "__main__":
    from rich.console import Console

    console = Console()
    setup_logging(level="TRACE")
    args = cli()



"""R2X API for data model."""

import csv
from collections.abc import Callable
from os import PathLike
from pathlib import Path
from itertools import chain
from collections.abc import Iterable
from loguru import logger

from infrasys.component import Component
from infrasys.system import System as ISSystem
from .__version__ import __data_model_version__
import uuid
import infrasys.cost_curves


class System(ISSystem):
    """API to interact with the SystemModel."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.data_format_version = __data_model_version__

    def __str__(self) -> str:
        return f"System(name={self.name}, DataModel Version={self.version})"

    def __repr__(self) -> str:
        return str(self)

    @property
    def version(self):
        """The version property."""
        return __data_model_version__

    def export_component_to_csv(
        self,
        component: type[Component],
        fields: list | None = None,
        filter_func: Callable | None = None,
        fpath: PathLike | None = None,
        key_mapping: dict | None = None,
        unnest_key: str = "name",
        **dict_writer_kwargs,
    ):
        """Export components into a csv.

        component:
            Component type to get from the system
        """
        # Get desired components to offload to csv
        components = map(
            lambda component: component.model_dump(
                exclude={"ext"}, exclude_none=True, mode="json", context={"magnitude_only": True}
            ),
            self.get_components(component, filter_func=filter_func),
        )
        if fpath is None:
            fpath = Path(f"{component.__name__}.csv")

        self._export_dict_to_csv(
            components,
            fpath=fpath,
            fields=fields,
            key_mapping=key_mapping,
            unnest_key=unnest_key,
            **dict_writer_kwargs,
        )

    def _add_operation_cost_data(  # noqa: C901
        self,
        data: Iterable[dict],
        fields: list | None = None,
    ):
        operation_cost_fields = set()
        for sub_dict in data:
            if "operation_cost" not in sub_dict.keys():
                continue

            operation_cost = sub_dict["operation_cost"]
            for cost_field_key, cost_field_value in operation_cost.items():
                if isinstance(cost_field_value, dict):
                    assert (
                        "uuid" in cost_field_value.keys()
                    ), f"Operation cost field {cost_field_key} was assumed to be a component but is not."
                    variable_cost = self.get_component_by_uuid(uuid.UUID(cost_field_value["uuid"]))
                    sub_dict["variable_cost"] = variable_cost.vom_units.function_data.proportional_term
                    if "fuel_cost" in variable_cost.model_fields:
                        # Note: We multiply the fuel price by 1000 to offset the division
                        # done by Sienna when it parses .csv files
                        sub_dict["fuel_price"] = variable_cost.fuel_cost * 1000
                        operation_cost_fields.add("fuel_price")

                    function_data = variable_cost.value_curve.function_data
                    if "constant_term" in function_data.model_fields:
                        sub_dict["heat_rate_a0"] = function_data.constant_term
                        operation_cost_fields.add("heat_rate_a0")
                    if "proportional_term" in function_data.model_fields:
                        sub_dict["heat_rate_a1"] = function_data.proportional_term
                        operation_cost_fields.add("heat_rate_a1")
                    if "quadratic_term" in function_data.model_fields:
                        sub_dict["heat_rate_a2"] = function_data.quadratic_term
                        operation_cost_fields.add("heat_rate_a2")
                    if "x_coords" in function_data.model_fields:
                        x_y_coords = dict(zip(function_data.x_coords, function_data.y_coords))
                        match type(variable_cost):
                            case infrasys.cost_curves.CostCurve:
                                for i, (x_coord, y_coord) in enumerate(x_y_coords.items()):
                                    output_point_col = f"output_point_{i}"
                                    sub_dict[output_point_col] = x_coord
                                    operation_cost_fields.add(output_point_col)

                                    cost_point_col = f"cost_point_{i}"
                                    sub_dict[cost_point_col] = y_coord
                                    operation_cost_fields.add(cost_point_col)

                            case infrasys.cost_curves.FuelCurve:
                                for i, (x_coord, y_coord) in enumerate(x_y_coords.items()):
                                    output_point_col = f"output_point_{i}"
                                    sub_dict[output_point_col] = x_coord
                                    operation_cost_fields.add(output_point_col)

                                    heat_rate_col = "heat_rate_avg_0" if i == 0 else f"heat_rate_incr_{i}"
                                    sub_dict[heat_rate_col] = y_coord
                                    operation_cost_fields.add(heat_rate_col)
                elif cost_field_key not in sub_dict.keys():
                    sub_dict[cost_field_key] = cost_field_value
                    operation_cost_fields.add(cost_field_key)
                else:
                    pass

        fields.remove("operation_cost")  # type: ignore
        fields.extend(list(operation_cost_fields))  # type: ignore

        return data, fields

    def _export_dict_to_csv(
        self,
        data: Iterable[dict],
        fpath: PathLike,
        fields: list | None = None,
        key_mapping: dict | None = None,
        unnest_key: str = "name",
        **dict_writer_kwargs,
    ):
        # Remaping keys
        # NOTE: It does not work recursively for nested components
        if key_mapping:
            data = [
                {key_mapping.get(key, key): value for key, value in sub_dict.items()} for sub_dict in data
            ]
            if fields:
                fields = list(map(lambda key: key_mapping.get(key, key), fields))

        if fields is None:
            fields = list(set(chain.from_iterable(data)))

        if "operation_cost" in fields:
            data, fields = self._add_operation_cost_data(data, fields)

        with open(str(fpath), "w", newline="") as csvfile:
            writer = csv.DictWriter(csvfile, fieldnames=fields, extrasaction="ignore", **dict_writer_kwargs)  # type: ignore
            writer.writeheader()
            for row in data:
                filter_row = {
                    key: value if not isinstance(value, dict) else value.get(unnest_key)
                    for key, value in row.items()
                }
                writer.writerow(filter_row)


if __name__ == "__main__":
    from .logger import setup_logging
    from rich.console import Console

    setup_logging(level="TRACE")
    logger.enable("infra_sys")

    console = Console()

    # from tests.models.systems import ieee_5bus
    #
    # system = ieee_5bus()
    # logger.info("From ARTEX")



"""Some util functions for the testing."""

from pathlib import Path
import shutil


def clean_folder(path: Path):
    """Remove all files from tmp folder."""
    if path.exists():
        shutil.rmtree(path)
    path.mkdir()



"""Exporter base class."""

from abc import ABC, abstractmethod
import string
from collections import defaultdict
from collections.abc import Callable
from pathlib import Path
from typing import Any

import pint
import pandas as pd
import numpy as np
import infrasys
from loguru import logger
from infrasys.base_quantity import BaseQuantity

from r2x.api import System
from r2x.config import Scenario
from r2x.parser.handler import file_handler

OUTPUT_FNAME = "{self.weather_year}"


class BaseExporter(ABC):
    """Class that defines the shared methods of parsers.

    Note
    ----
    This class is meant to be use for developing new parsers. Do not use it directly.

    Attributes
    ----------
    config:
        Scenario configuration
    system:
        System class
    output_foder:
        output folder path

    Methods
    -------
    run()
        Sequentially run methods from the exporter
    get_data(key='load')
        Return the parsed data for load.
    read_file(fpath="load.csv")
        Read load data.
    export_data_files(time_series_folder="Data")
        Save time series data files.
    """

    def __init__(
        self, config: Scenario, system: System, output_folder: Path, ts_directory: str = "Data"
    ) -> None:
        self.config = config
        self.system = system
        self.output_folder = output_folder
        self.input_model = config.input_model
        self.ts_directory = Path(ts_directory)
        self.time_series_objects: dict[str, list[Any]] = defaultdict(list)
        self.time_series_name_by_type: dict[str, list[Any]] = defaultdict(list)
        self._handle_data_folder(config.output_folder, self.ts_directory)

    def _handle_data_folder(self, output_folder: str | Path, folder_name: str | Path) -> None:
        fpath = Path(output_folder) / folder_name
        if not fpath.exists():
            fpath.mkdir()
        return

    @abstractmethod
    def run(self, *args, **kwargs) -> "BaseExporter":
        """Run the exporter methods in sequence.

        This methods needs to be implemented by any exporter and it's content
        change based on the order the data is being exported.
        """

    def read_file(self, fpath: Path | str, filter_func: list[Callable] | None = None, **kwargs):
        """Read input model data from the file system.

        Currently supported formats:
            - .csv
            - .h5
        More to come!

        Parameters
        ----------
        fpath: Path, str
            Absolute location of the file in the system
        filter_func: List
            Filter functions to apply

        """
        data = file_handler(fpath, **kwargs)
        if data is None:
            return

        if isinstance(filter_func, list):
            for func in filter_func:
                data = func(data, **kwargs)
        return data

    def export_data_files(self, time_series_folder: str = "Data") -> None:
        """Export all time series objects attached to components.

        This method assumes that `self.config.weather_year and `self.output_folder` exist.

        Parameters
        ----------
        time_series_folder: str
            Folder name to save time series data
        """
        config_dict = self.config.__dict__
        for component in self.system.iter_all_components():
            if self.system.has_time_series(component):
                for ts_metadata in self.system.time_series.list_time_series_metadata(component):
                    ts_component_name = f"{component.__class__.__name__}_{ts_metadata.variable_name}"
                    self.time_series_objects[ts_component_name].append(
                        self.system.get_time_series(component, variable_name=ts_metadata.variable_name)
                    )
                    self.time_series_name_by_type[ts_component_name].append(component.name)

        assert self.config.weather_year is not None
        date_time_column = pd.date_range(
            start=f"1/1/{self.config.weather_year}",
            end=f"1/1/{self.config.weather_year + 1}",
            freq="1h",
            inclusive="left",
        )
        date_time_column = np.datetime_as_string(date_time_column, unit="m")
        # Remove leap day to match ReEDS convention
        # date_time_column = date_time_column[~((date_time_column.month == 2) & (date_time_column.day == 29))]
        if self.input_model == "reeds-US":
            date_time_column = date_time_column[:-24]

        csv_fpath = self.output_folder / time_series_folder

        # Use string substitution to dynamically change the output csv fnames
        csv_fname = config_dict.get("time_series_fname", "${component_type}_${name}_${weather_year}.csv")
        string_template = string.Template(csv_fname)

        for component_type, time_series in self.time_series_objects.items():
            time_series_arrays = list(map(lambda x: x.data.to_numpy(), time_series))

            config_dict["component_type"] = component_type
            csv_fname = string_template.safe_substitute(config_dict)
            csv_table = np.column_stack([date_time_column, *time_series_arrays])
            header = '"DateTime",' + ",".join(
                [f'"{name}"' for name in self.time_series_name_by_type[component_type]]
            )

            np.savetxt(
                csv_fpath / csv_fname,
                csv_table,
                header=header,
                delimiter=",",
                comments="",
                fmt="%s",
            )

        return

    def get_valid_records_properties(
        self,
        component_list,
        property_map: dict[str, str],
        unit_map: dict[str, str],
        valid_properties: list | None = None,
    ):
        """Return a validadted list of properties to the given property_map."""
        result = []
        component_list_mapped = [
            {property_map.get(key, key): value for key, value in d.items()} for d in component_list
        ]
        for component in component_list_mapped:
            component_dict = {"name": component["name"]}  # We need the name to match it with the membership.
            for property_name, property_value in component.items():
                if valid_properties is not None:
                    if property_name in valid_properties:
                        property_value = self.get_property_magnitude(
                            property_value, to_unit=unit_map.get(property_name)
                        )
                        component_dict[property_name] = property_value
                else:
                    property_value = self.get_property_magnitude(
                        property_value, to_unit=unit_map.get(property_name)
                    )
                    component_dict[property_name] = property_value
            result.append(component_dict)
        return result

    def get_property_magnitude(self, property_value, to_unit: str | None = None) -> float:
        """Return magnitude with the given units for a pint Quantity.

        Parameters
        ----------
        property_name

        property_value
            pint.Quantity to extract magnitude from
        to_unit
            String that contains the unit conversion desired. Unit must be compatible.
        """
        if not isinstance(property_value, pint.Quantity | BaseQuantity):
            return property_value
        if to_unit:
            unit = to_unit.replace("$", "usd")  # Dollars are named usd on pint
            property_value = property_value.to(unit)
        return property_value.magnitude


def get_exporter(
    config: Scenario,
    system: infrasys.system.System,
    exporter_class,
    filter_funcs: list[Callable] | None = None,
    **kwargs,
) -> BaseExporter:
    """Return exporter instance.


    Parameters
    ----------
    config
        Scenario configuration class
    system
        Infrasys system
    filter_func:
        Functions that will applied to read_data process

    Other Parameters
    ----------------
    kwargs
        year
            For filtering by solve year
        year_column
            To change the column to apply the filter
        column_mapping
            For renaming columns

    See Also
    --------
    BaseExporter
    pl_filter_year
    pl_lower_case
    pl_rename
    """
    logger.debug("Getting {} instance.", exporter_class.__name__)

    exporter: BaseExporter = exporter_class(
        config=config, system=system, output_folder=config.output_folder, **kwargs
    )

    # Functions relative to the parser.
    # NOTE: At some point we are going to migrate this out, but this sound like a good standard set
    # if filter_funcs is None:
    #     logger.trace("Using default filter functions")
    #     filter_funcs = [pl_lowercase, pl_rename, pl_filter_year]

    exporter.run(
        config=config,
        fmap=config.fmap,
        year=config.solve_year,
        filter_func=filter_funcs,
        **kwargs,
    )

    return exporter



"""Script that creates simple 2-area pjm systems for testing."""

from datetime import datetime, timedelta
import pathlib

from infrasys.time_series_models import SingleTimeSeries
from r2x.api import System
from r2x.enums import ACBusTypes, PrimeMoversType
from r2x.models.branch import AreaInterchange, Line, MonitoredLine
from r2x.models.generators import RenewableDispatch, ThermalStandard
from r2x.models.load import PowerLoad
from r2x.models.topology import ACBus, Area, LoadZone
from r2x.units import ActivePower, Percentage, Time, Voltage, ureg
from r2x.utils import read_json


def pjm_2area() -> System:
    """Return the PJM 2-area test system."""
    fpath = pathlib.Path(__file__).parent.parent  # So it points to tests
    fname = "data/pjm_2area_data.json"
    pjm_2area_components = read_json(str(fpath / fname))

    system = System(
        name="pjm 2-area system", auto_add_composed_components=True, description="Test system for PJM"
    )

    # Add topology elements
    system.add_component(Area(name="init"))
    system.add_component(Area(name="Area2"))
    system.add_component(LoadZone(name="LoadZone1"))

    for bus in pjm_2area_components["bus"]:
        system.add_component(
            ACBus(
                name=bus["name"],
                number=bus["number"],
                bus_type=ACBusTypes(bus["bustype"]),
                base_voltage=Voltage(bus["base_voltage"], "kV"),
                area=system.get_component(Area, bus["area"]),
                magnitude=bus["magnitude"],
            )
        )

    # Add branches
    for branch in pjm_2area_components["branch"]:
        busf = system.get_component(ACBus, branch["FromBus"])
        bust = system.get_component(ACBus, branch["ToBus"])
        system.add_component(
            Line(
                name=branch["Name"],
                from_bus=busf,
                to_bus=bust,
                x=branch["x"],
                r=branch["r"],
                rating=branch["MaxRating"] * ureg.MW,
            )
        )
    # Add MonitoredLine
    busf = system.get_component(ACBus, "Bus_nodeC_1")
    bust = system.get_component(ACBus, "Bus_nodeC_2")
    branch_monitored = MonitoredLine(
        name="inter_area_line",
        from_bus=busf,
        r=0.003,
        x=0.03,
        to_bus=bust,
        rating_up=1000 * ureg.MW,
        rating=1000 * ureg.MW,
    )
    system.add_component(branch_monitored)

    # Add area interchange
    system.add_component(
        AreaInterchange(
            name="1_2",
            max_power_flow=150.0 * ureg.MW,
            min_power_flow=-150.0 * ureg.MW,
            from_area=system.get_component(Area, "init"),
            to_area=system.get_component(Area, "Area2"),
        )
    )

    # Add thermal generators
    for gen in pjm_2area_components["thermal"]:
        system.add_component(
            ThermalStandard(
                name=gen["Name"],
                fuel=gen["fuel"].lower(),
                prime_mover_type=PrimeMoversType.ST,
                unit_type=PrimeMoversType.ST,
                rating=gen["Rating"] * ureg.MVA,
                active_power=ActivePower(100, "MW"),
                min_rated_capacity=gen["Pmin"] * ureg.MW,
                startup_cost=gen["StartupCost"] * ureg.Unit("usd"),
                shutdown_cost=gen["ShutDnCost"] * ureg.Unit("usd"),
                vom_price=gen["VOM"] * ureg.Unit("usd/MWh"),
                min_down_time=gen["MinTimeDn"],
                min_up_time=gen["MinTimeUp"],
                mean_time_to_repair=Time(10.0, "hour"),
                forced_outage_rate=Percentage(0.0),
                planned_outage_rate=Percentage(0.0),
                ramp_up=gen["RampLimitsUp"],
                ramp_down=gen["RampLimitsDn"],
                bus=system.get_component(ACBus, gen["BusName"]),
            )
        )

    # Solar generators
    solar_pv_01 = RenewableDispatch(
        name="PVBus5",
        bus=system.get_component(ACBus, "Bus_nodeC_1"),
        prime_mover_type=PrimeMoversType.PV,
        unit_type=PrimeMoversType.PV,
        active_power=384 * ureg.MW,
        category="solar",
    )
    system.add_component(solar_pv_01)

    wind_01 = RenewableDispatch(
        name="WindBus1",
        bus=system.get_component(ACBus, "Bus_nodeA_2"),
        prime_mover_type=PrimeMoversType.WT,
        unit_type=PrimeMoversType.WT,
        active_power=451.0 * ureg.MW,
        category="wind",
    )
    system.add_component(wind_01)

    # Add renewable profiles
    initial_time = datetime(year=2024, month=1, day=1)
    solar_array = pjm_2area_components["solar_ts"]  # Data is hourly resolution
    wind_array = pjm_2area_components["wind_ts"]  # Data is hourly resolution
    solar_ts = SingleTimeSeries.from_array(
        data=solar_array,
        initial_time=initial_time,
        resolution=timedelta(hours=1),
        variable_name="max_active_power",
    )
    wind_ts = SingleTimeSeries.from_array(
        data=wind_array,
        initial_time=initial_time,
        resolution=timedelta(hours=1),
        variable_name="rated_capacity",
    )
    #
    system.add_time_series(solar_ts, solar_pv_01)
    system.add_time_series(wind_ts, wind_01)

    initial_time = datetime(year=2024, month=1, day=1)
    for load in pjm_2area_components["load"]:
        load_component = PowerLoad(
            name=load["Name"],
            bus=system.get_component(ACBus, load["BusName"]),
            max_active_power=load["MaxLoad"] * ureg.MW,
        )
        ld_ts = SingleTimeSeries.from_array(
            data=load["ts"],
            initial_time=initial_time,
            resolution=timedelta(hours=1),
            variable_name="max_active_power",
        )
        system.add_component(load_component)
        system.add_time_series(ld_ts, load_component)
    return system



"""R2X pytest fixtures.

Here goes all the variables that will be shared between all the different testing scripts.
"""

import pytest
from r2x.utils import read_json
from loguru import logger
from _pytest.logging import LogCaptureFixture


DATA_FOLDER = "tests/data"
OUTPUT_FOLDER = "r2x_output"


@pytest.fixture
def data_folder(pytestconfig):
    return pytestconfig.rootpath.joinpath(DATA_FOLDER)


@pytest.fixture
def tmp_folder(tmp_path_factory):
    tmp_path = tmp_path_factory.mktemp(OUTPUT_FOLDER)
    yield tmp_path


@pytest.fixture
def defaults_dict() -> dict[str, str]:
    config_dict = read_json("r2x/defaults/config.json")
    plugins_dict = read_json("r2x/defaults/plugins_config.json")
    plexos_dict = read_json("r2x/defaults/plexos_output.json")
    sienna_dict = read_json("r2x/defaults/sienna_config.json")
    combined_dict = config_dict | plugins_dict | plexos_dict | sienna_dict
    return combined_dict


@pytest.fixture
def caplog(caplog: LogCaptureFixture):
    handler_id = logger.add(
        caplog.handler,
        format="{message}",
        level=0,
        filter=lambda record: record["level"].no >= caplog.handler.level,
        enqueue=False,  # Set to 'True' if your test is spawning child processes.
    )
    yield caplog
    logger.remove(handler_id)



"""Models that capture topology types."""

from typing import Annotated

from pydantic import Field, NonNegativeFloat, PositiveFloat, PositiveInt

from r2x.enums import ACBusTypes
from r2x.models.core import BaseComponent
from r2x.units import Voltage, ureg


class Topology(BaseComponent):
    """Abstract type to represent the structure and interconnectedness of the system."""


class AggregationTopology(Topology):
    """Base class for area-type components."""


class Area(AggregationTopology):
    """Collection of buses in a given region."""

    peak_active_power: Annotated[NonNegativeFloat, Field(description="Peak active power in the area")] = 0.0
    peak_reactive_power: Annotated[NonNegativeFloat, Field(description="Peak reactive power in the area")] = (
        0.0
    )
    load_response: Annotated[
        float,
        Field(
            description=(
                "Load-frequency damping parameter modeling how much the load in the area changes "
                "due to changes in frequency (MW/Hz)."
            )
        ),
    ] = 0.0

    @classmethod
    def example(cls) -> "Area":
        return Area(name="New York")


class LoadZone(AggregationTopology):
    """Collection of buses for electricity price analysis."""

    @classmethod
    def example(cls) -> "LoadZone":
        return LoadZone(name="ExampleLoadZone")


class Bus(Topology):
    """Abstract class for a bus."""

    number: Annotated[PositiveInt, Field(description="A unique bus identification number.")]
    bus_type: Annotated[ACBusTypes, Field(description="Type of category of bus,")] | None = None
    area: Annotated[Area, Field(description="Area containing the bus.")] | None = None
    load_zone: Annotated[LoadZone, Field(description="the load zone containing the DC bus.")] | None = None
    base_voltage: (
        Annotated[Voltage, Field(gt=0, description="Base voltage in kV. Unit compatible with voltage.")]
        | None
    ) = None
    magnitude: (
        Annotated[PositiveFloat, Field(description="Voltage as a multiple of base_voltage.")] | None
    ) = None


class DCBus(Bus):
    """Power-system DC Bus."""

    @classmethod
    def example(cls) -> "DCBus":
        return DCBus(
            number=1,
            name="ExampleBus",
            load_zone=LoadZone.example(),
            area=Area.example(),
            base_voltage=20 * ureg.kV,
            bus_type=ACBusTypes.PV,
        )


class ACBus(Bus):
    """Power-system AC bus."""

    angle: Annotated[float | None, Field(description="Angle of the bus in radians.", gt=-1.571, lt=1.571)] = (
        None
    )

    @classmethod
    def example(cls) -> "ACBus":
        return ACBus(
            number=1,
            name="ExampleBus",
            load_zone=LoadZone.example(),
            area=Area.example(),
            base_voltage=13 * ureg.kV,
            bus_type=ACBusTypes.PV,
        )



"""R2X Sienna system exporter."""

# System packages
import json
import os
from typing import Any

# Third-party packages
from loguru import logger

# Local imports
from r2x.exporter.handler import BaseExporter
from r2x.models import (
    ACBranch,
    Bus,
    DCBranch,
    Generator,
    HydroPumpedStorage,
    PowerLoad,
    Reserve,
    ReserveMap,
    Storage,
)


class SiennaExporter(BaseExporter):
    """Sienna exporter class.

    This is the class that export the infrasys system to csv's required to read
    in PowerSystems.jl. The class contains a method per file that needs to get
    created and some of the methods manipulate the data to be compliant wth
    PowerSystems.jl. The order is important for some steps, however, most
    of them could be run sequentially.

    The method `run` is the responsible of running all the exporter.

    Notes
    -----
    All the methods defined an `output_columns` which defines the fields from
    the `Component` that we want to export.

    All methods call `export_component_to_csv` or `export_dict_to_csv` that
    is defined on the `BaseExporter` class.


    See Also
    --------
    BaseExporter.export_component_to_csv
    BaseExporter.export_dict_to_csv
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.output_data = {}
        self.property_map = self.config.defaults.get("sienna_property_map", {})

    def run(self, *args, path=None, **kwargs) -> "SiennaExporter":
        """Run sienna exporter workflow.


        Notes
        -----
        All methods call `export_component_to_csv` or `export_dict_to_csv` that
        is defined on the `BaseExporter` class.

        If performance is a constraint, we could parallelize the methods.
        """
        logger.info("Starting {}", self.__class__.__name__)

        self.process_bus_data()
        self.process_load_data()
        self.process_branch_data()
        self.process_dc_branch_data()
        self.process_gen_data()
        self.process_reserves_data()
        self.process_storage_data()
        self.export_data()
        self.create_timeseries_pointers()
        return self

    def process_bus_data(self, fname: str = "bus.csv") -> None:
        """Create bus.csv file.

        Parameters
        ----------
        fname : str
            Name of the file to be created
        """
        output_fields = [
            "bus_id",
            "name",
            "area",
            "zone",
            "base_voltage",
            "bus_type",
        ]
        self.system.export_component_to_csv(
            Bus,
            fpath=self.output_folder / fname,
            fields=output_fields,
            key_mapping={"number": "bus_id", "load_zone": "zone"},
            restval="NA",
        )

    def process_load_data(self, fname: str = "load.csv") -> None:
        """Create load.csv file.

        Parameters
        ----------
        fname : str
            Name of the file to be created
        """
        output_fields = [
            "bus_id",
            "name",
            "available",
            "active_power",
            "reactive_power",
            "max_active_power",
            "max_reeactive_power",
            "active_power",
        ]
        self.system.export_component_to_csv(
            PowerLoad,
            fpath=self.output_folder / fname,
            fields=output_fields,
            unnest_key="number",
            key_mapping={"bus": "bus_id"},
            restval="0.0",
        )
        logger.info(f"File {fname} created.")

    def process_branch_data(self, fname: str = "branch.csv") -> None:
        """Create branch.csv file.

        Parameters
        ----------
        fname : str
            Name of the file to be created
        """
        output_fields = [
            "name",
            "connection_points_from",
            "connection_points_to",
            "r",
            "x",
            "b",
            "rate",
            "branch_type",
        ]

        # NOTE: We need to decide what we do if the user provides a rate or bi-directional rate
        # if "rate" in output_df.columns:
        #     output_df["rate"] = output_df["rate"].fillna(
        #         (output_df["rating_up"] + np.abs(output_df["rating_down"])) / 2
        #     )
        # else:
        #     output_df["rate"] = (output_df["rating_up"] + np.abs(output_df["rating_down"])) / 2

        self.system.export_component_to_csv(
            ACBranch,
            fpath=self.output_folder / fname,
            fields=output_fields,
            unnest_key="number",
            key_mapping={
                "from_bus": "connection_points_from",
                "to_bus": "connection_points_to",
                "class_type": "branch_type",
                "rating": "rate",
                "b": "primary_shunt",
            },
            # restval=0.0,
        )
        logger.info(f"File {fname} created.")

    def process_dc_branch_data(self, fname="dc_branch.csv") -> None:
        """Create dc_branch.csv file.

        Parameters
        ----------
        fname : str
            Name of the file to be created
        """
        output_fields = [
            "name",
            "connection_points_from",
            "connection_points_to",
            "rate",
            "loss",
        ]

        # NOTE: We need to decide what we do if the user provides a rate or bi-directional rate
        # if "rate" in output_df.columns:
        #     output_df["rate"] = output_df["rate"].fillna(
        #         (output_df["rating_up"] + np.abs(output_df["rating_down"])) / 2
        #     )
        # else:
        #     output_df["rate"] = (output_df["rating_up"] + np.abs(output_df["rating_down"])) / 2
        self.system.export_component_to_csv(
            DCBranch,
            fpath=self.output_folder / fname,
            fields=output_fields,
            unnest_key="number",
            key_mapping={
                "from_bus": "connection_points_from",
                "to_bus": "connection_points_to",
                "class_type": "branch_type",
                "rating_up": "rate",
            },
            restval="NA",
        )
        logger.info(f"File {fname} created.")
        return

    def process_gen_data(self, fname="gen.csv"):
        """Create gen.csv file.

        Parameters
        ----------
        fname : str
            Name of the file to be created
        """
        output_fields = [
            "name",
            "available",
            "prime_mover_type",
            "bus_id",
            "fuel",
            "base_mva",
            "rating",
            "unit_type",
            "active_power",
            "min_rated_capacity",
            "min_down_time",
            "min_up_time",
            "mean_time_to_repair",
            "forced_outage_rate",
            "planned_outage_rate",
            "ramp_up",
            "ramp_down",
            "category",
            "must_run",
            "pump_load",
            "vom_price",
            "operation_cost",
        ]

        key_mapping = {"bus": "bus_id"}
        self.system.export_component_to_csv(
            Generator,
            fpath=self.output_folder / fname,
            fields=output_fields,
            key_mapping=key_mapping,
            unnest_key="number",
            restval="NA",
        )
        logger.info(f"File {fname} created.")

    def process_reserves_data(self, fname="reserves.csv") -> None:
        """Create reserve.csv file.

        Parameters
        ----------
        fname : str
            Name of the file to be created
        """
        output_fields = [
            "name",
            "requirement",
            "timeframe",
            "eligible_region",
            "direction",
            "contributing_devices",
            "eligible_device_categories",
            "eligible_device_subcategories",
        ]

        # ReserveMap holds the eligible devices in the mapping.
        reserve_map_list: list[dict[str, Any]] = list(
            map(lambda component: component["mapping"], self.system.to_records(ReserveMap))
        )

        if not reserve_map_list:
            msg = "Reserve map class not found on the system. Skipping reserve contributing devices file."
            logger.warning(msg)
            return

        if len(reserve_map_list) > 1:
            logger.warning("We do not support multiple reserve maps per system")
            return
        reserve_map: dict = reserve_map_list[0]

        reserves: list[dict[str, Any]] = list(self.system.to_records(Reserve))

        output_data = []
        for reserve in reserves:
            # NOTE: Only export reserves that exists on the reserve map
            if reserve["name"] in reserve_map:
                output_dict = reserve
                output_dict["direction"] = reserve["direction"].name
                output_dict["eligible_device_categories"] = "(Generator,Storage)"
                contributing_devices = reserve_map.get(reserve["name"])
                output_dict["contributing_devices"] = str(tuple(contributing_devices)).replace(  # type: ignore
                    "'", ""
                )  # Sienna expects a tuple for this field that looks like "(val1,val2,val3)"
                output_data.append(output_dict)

        key_mapping = {"region": "eligible_region", "max_requirement": "requirement"} | self.property_map
        self.system._export_dict_to_csv(
            output_data,
            fpath=self.output_folder / fname,
            fields=output_fields,
            key_mapping=key_mapping,
            unnest_key="name",
            restval="NA",
        )
        logger.info(f"File {fname} created.")
        return

    def process_storage_data(self, fname="storage.csv") -> None:
        """Create storage.csv file.

        Parameters
        ----------
        fname : str
            Name of the file to be created
        """
        output_fields = [
            "name",
            "position",
            "available",
            "generator_name",
            "bus_id",
            "active_power",
            "rating",
            "input_efficiency",
            "output_efficiency",
            "storage_capacity",
            "min_storage_capacity",
            "max_storage_capacity",
            "input_active_power_limit_min",
            "output_active_power_limit_max",
            "output_active_power_limit_min",
            "unit_type",
        ]

        generic_storage = self.get_valid_records_properties(
            self.system.to_records(Storage),
            property_map=self.config.defaults["sienna_property_map"],
            unit_map=self.config.defaults["sienna_unit_map"],
            # valid_properties=output_fields,
        )
        hydro_pump = list(self.system.to_records(HydroPumpedStorage))
        storage_list = generic_storage + hydro_pump

        if storage_list is None:
            logger.warning("No storage devices found")
            return

        output_data = []
        for storage in storage_list:
            output_dict = storage
            output_dict["generator_name"] = storage["name"]
            output_dict["input_active_power_limit_max"] = output_dict["active_power"]
            output_dict["output_active_power_limit_max"] = output_dict["active_power"]
            # NOTE: If we need to change this in the future, we could probably
            # use the function max to check if the component has the field.
            output_dict["input_active_power_limit_min"] = 0  # output_dict["active_power"]
            output_dict["output_active_power_limit_min"] = 0  # output_dict["active_power"]
            output_dict["active_power"] = output_dict["active_power"]
            output_dict["bus_id"] = getattr(self.system.get_component_by_label(output_dict["bus"]), "number")
            output_dict["rating"] = output_dict["rating"]

            # NOTE: For pumped hydro storage we create a head and a tail
            # representation that keeps track of the upper and down reservoir
            # state of charge.
            # if storage["class_type"] == "HydroPumpedStorage":
            original_name = output_dict["name"]
            output_dict["name"] = original_name + "_head"
            output_dict["position"] = "head"
            output_data.append(output_dict)
            tail_copy = output_dict.copy()
            tail_copy["name"] = original_name + "_tail"
            output_dict["position"] = "tail"
            output_data.append(tail_copy)

        key_mapping = self.property_map
        self.system._export_dict_to_csv(
            output_data,
            fpath=self.output_folder / fname,
            fields=output_fields,
            key_mapping=key_mapping,
            unnest_key="number",
            restval="NA",
        )

        logger.info("File storage.csv created.")

    def create_timeseries_pointers(self) -> None:
        """Create timeseries_pointers.json file.

        Parameters
        ----------
        fname : str
            Name of the file to be created
        """
        ts_pointers_list = []

        for component_type, time_series in self.time_series_objects.items():
            csv_fpath = self.ts_directory / (
                f"{component_type}_{self.config.name}_{self.config.weather_year}.csv"
            )
            for i in range(len(time_series)):
                component_name = self.time_series_name_by_type[component_type][i]
                ts_instance = time_series[i]
                resolution = ts_instance.resolution.seconds
                variable_name = self.property_map.get(ts_instance.variable_name, ts_instance.variable_name)
                # TODO(pedro): check if the time series data is pre normalized
                # https://github.nrel.gov/PCM/R2X/issues/417
                ts_pointers = {
                    "category": component_type.split("_", maxsplit=1)[0],  # Component_name is the first
                    "component_name": component_name,
                    "data_file": str(csv_fpath),
                    "normalization_factor": 1.0,
                    "resolution": resolution,
                    "name": variable_name,
                    "scaling_factor_multiplier_module": None,
                    "scaling_factor_multiplier": None,
                }
                ts_pointers_list.append(ts_pointers)

        with open(os.path.join(self.output_folder, "timeseries_pointers.json"), mode="w") as f:
            json.dump(ts_pointers_list, f)

        logger.info("File timeseries_pointers.json created.")
        return

    def export_data(self) -> None:
        """Export csv data to specified folder from output_data attribute."""
        logger.debug("Saving Sienna data and timeseries files.")

        # First export all time series objects
        self.export_data_files()
        logger.info("Saving time series data.")



"""Test that we can read defaults data files and configurations."""

import json
import pytest
import importlib.resources as resources

DEFAULTS_PATH = "r2x.defaults"
DEFAULT_MAPS_NAME = ["plexos_input.json", "sienna_config.json"]


@pytest.fixture(scope="module")
def json_data() -> list[dict[str, str]]:
    loaded_data = []
    file_paths = [resources.files(DEFAULTS_PATH).joinpath(f) for f in DEFAULT_MAPS_NAME]
    for path in file_paths:
        with path.open() as file:
            loaded_data.append(json.load(file))
    return loaded_data


def test_json_data(json_data: list[dict[str, str]]) -> None:
    assert len(json_data) > 0
    for data in json_data:
        assert isinstance(data, dict)
        assert len(data) > 0



"""R2X logger configuration."""

# System packages
import argparse
import os
import sys

# Third-party packages
from loguru import logger

# Logger printing formats
DEFAULT_FORMAT = "<level>{level}</level>: {message}"


class Formatter:  # noqa: D101
    def __init__(self):
        self.padding = 0
        self.fmt = "<green>{time:YYYY-MM-DD HH:mm:ss}</green> | <level>{level: <8}</level>| <cyan>{name}:{line}{extra[padding]}</cyan> | {message}\n{exception}"  # noqa: E501

    def format(self, record):  # noqa: D102
        length = len("{name}:{line}".format(**record))
        self.padding = max(self.padding, length)
        record["extra"]["padding"] = " " * (self.padding - length)
        return self.fmt


def setup_logging(
    filename=None,
    level="INFO",
    debug: bool = False,
):
    """Configure logging of file.

    Parameters
    ----------
    filename : str | None
        log filename
    level : str, optional
        change defualt level of logging.
    verbose :  bool
        returns additional logging information.
    """
    if debug:
        level = "DEBUG"
    from loguru import logger

    logger.remove()
    logger.enable("r2x")
    # logger.enable("infrasys")
    # logger.enable("resource_monitor")
    level = os.environ["LOGURU_LEVEL"] if os.environ.get("LOGURU_LEVEL") else level
    logger.add(
        sys.stderr,
        level=level,
        enqueue=False,
        format=Formatter().format if level == "DEBUG" or level == "TRACE" else DEFAULT_FORMAT,
    )
    if filename:
        logger.add(filename, level=level, enqueue=True)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--debug", action="store_true")
    args = parser.parse_args()

    setup_logging(level="TRACE")

    # Testing different types of loggers.
    logger.trace("A trace message.")
    logger.debug("A debug message.")
    logger.info("An info message.")
    logger.success("A success message.")
    logger.warning("A warning message.")
    logger.error("An error message.")
    logger.critical("A critical message.")



"""Core models for R2X."""

from collections import defaultdict, namedtuple

from infrasys.component import Component
from typing import Annotated
from pydantic import Field, field_serializer
from r2x.units import ureg


class BaseComponent(Component):
    """Infrasys base component with additional fields for R2X."""

    available: Annotated[bool, Field(description="If the component is available.")] = True
    category: Annotated[str, Field(description="Category that this component belongs to.")] | None = None
    ext: dict = Field(default_factory=dict, description="Additional information of the component.")

    @property
    def class_type(self) -> str:
        """Create attribute that holds the class name."""
        return type(self).__name__

    @field_serializer("ext", when_used="json")
    def serialize_ext(ext: dict):  # type:ignore  # noqa: N805
        for key, value in ext.items():
            if isinstance(value, ureg.Quantity):
                ext[key] = value.magnitude
        return ext


MinMax = namedtuple("MinMax", ["min", "max"])


class Service(BaseComponent):
    """Abstract class for services attached to components."""


class Device(BaseComponent):
    """Abstract class for devices."""

    services: (
        Annotated[
            list[Service],
            Field(description="Services that this component contributes to.", default_factory=list),
        ]
        | None
    ) = None


class StaticInjection(Device):
    """Supertype for all static injection devices."""


class TransmissionInterfaceMap(BaseComponent):
    mapping: defaultdict[str, list] = defaultdict(list)  # noqa: RUF012


class ReserveMap(BaseComponent):
    mapping: defaultdict[str, list] = defaultdict(list)  # noqa: RUF012



"""Functions related to parsers."""

from datetime import datetime, timedelta
import importlib
from importlib.resources import files
from pathlib import Path, PureWindowsPath
from argparse import ArgumentParser
import pandas as pd
import re

from pint import UndefinedUnitError, Quantity
import polars as pl
import numpy as np
from loguru import logger
from infrasys.exceptions import ISNotStored
from infrasys.time_series_models import SingleTimeSeries
from infrasys.value_curves import InputOutputCurve, AverageRateCurve
from infrasys.cost_curves import FuelCurve
from infrasys.function_data import (
    LinearFunctionData,
    QuadraticFunctionData,
)

from r2x.units import ureg
from r2x.api import System
from r2x.config import Scenario
from r2x.enums import ACBusTypes, ReserveDirection, ReserveType, PrimeMoversType
from r2x.exceptions import ModelError, ParserError
from plexosdb import PlexosSQLite
from plexosdb.enums import ClassEnum, CollectionEnum
from r2x.models.costs import ThermalGenerationCost, RenewableGenerationCost, HydroGenerationCost
from r2x.models import (
    ACBus,
    Generator,
    GenericBattery,
    MonitoredLine,
    PowerLoad,
    Reserve,
    LoadZone,
    ReserveMap,
    TransmissionInterface,
    TransmissionInterfaceMap,
    RenewableGen,
    ThermalGen,
    HydroDispatch,
)
from r2x.utils import validate_string

from .handler import PCMParser
from .parser_helpers import (
    handle_leap_year_adjustment,
    fill_missing_timestamps,
    resample_data_to_hourly,
    filter_property_dates,
    field_filter,
    prepare_ext_field,
)

models = importlib.import_module("r2x.models")

R2X_MODELS = importlib.import_module("r2x.models")
BASE_WEATHER_YEAR = 2007
XML_FILE_KEY = "xml_file"
PROPERTY_SV_COLUMNS_BASIC = ["name", "value"]
PROPERTY_SV_COLUMNS_NAMEYEAR = ["name", "year", "month", "day", "period", "value"]
PROPERTY_TS_COLUMNS_BASIC = ["year", "month", "day", "period", "value"]
PROPERTY_TS_COLUMNS_MULTIZONE = ["year", "month", "day", "period"]
PROPERTY_TS_COLUMNS_PIVOT = ["year", "month", "day"]
PROPERTY_TS_COLUMNS_YM = ["year", "month"]
PROPERTY_TS_COLUMNS_MDP = ["month", "day", "period"]
PROPERTY_TS_COLUMNS_MONTH_PIVOT = [
    "name",
    "m01",
    "m02",
    "m03",
    "m04",
    "m05",
    "m06",
    "m07",
    "m08",
    "m09",
    "m10",
    "m11",
    "m12",
]
DEFAULT_QUERY_COLUMNS_SCHEMA = {  # NOTE: Order matters
    "membership_id": pl.Int64,
    "parent_object_id": pl.Int32,
    "parent_object_name": pl.String,
    "parent_class_name": pl.String,
    "child_class_name": pl.String,
    "category": pl.String,
    "object_id": pl.Int32,
    "name": pl.String,
    "property_name": pl.String,
    "property_unit": pl.String,
    "property_value": pl.Float32,
    "band": pl.Int32,
    "date_from": pl.String,
    "date_to": pl.String,
    "memo": pl.String,
    "scenario_category": pl.String,
    "scenario": pl.String,
    "action": pl.String,
    "data_file_tag": pl.String,
    "data_file": pl.String,
    "variable_tag": pl.String,
    "timeslice_tag": pl.String,
    "timeslice": pl.String,
    "timeslice_value": pl.Float32,
}
COLUMNS = [
    "name",
    "property_name",
    "property_value",
    "band",
    "scenario",
    "date_from",
    "date_to",
    "text",
]
DEFAULT_INDEX = ["object_id", "name", "category"]


def cli_arguments(parser: ArgumentParser):
    """CLI arguments for the plugin."""
    parser.add_argument(
        "--model",
        required=False,
        help="Plexos model to translate",
    )


class PlexosParser(PCMParser):
    """Plexos parser class."""

    def __init__(self, *args, xml_file: str | None = None, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        assert self.config.run_folder
        self.run_folder = Path(self.config.run_folder)
        self.system = System(name=self.config.name, auto_add_composed_components=True)
        self.property_map = self.config.defaults["plexos_property_map"]
        self.device_map = self.config.defaults["plexos_device_map"]
        self.fuel_map = self.config.defaults["plexos_fuel_map"]
        self.device_match_string = self.config.defaults["device_name_inference_map"]

        # TODO(pesap): Rename exceptions to include R2X
        # https://github.com/NREL/R2X/issues/5
        # R2X needs at least one of this maps defined to correctly work.
        if not self.fuel_map and not self.device_map and not self.device_match_string:
            msg = (
                "Neither `plexos_fuel_map` or `plexos_device_map` or `device_match_string` was provided. "
                "To fix, provide any of the mappings."
            )
            raise ParserError(msg)

        # Populate databse from XML file.
        xml_file = xml_file or self.run_folder / self.config.fmap["xml_file"]["fname"]
        self.db = PlexosSQLite(xml_fname=xml_file)

        # Extract scenario data
        model_name = getattr(self.config, "model", None)
        logger.debug("Parsing plexos model={}", model_name)
        if model_name is None:
            model_name = self._select_model_name()
        self._process_scenarios(model_name=model_name)

        # date from is in days since 1900, convert to year
        date_from = self._collect_horizon_data(model_name=model_name).get("Date From")
        if date_from is not None:
            self.study_year: int = int((date_from / 365.25) + 1900)
            self.config.weather_year = self.study_year
        else:
            if self.config.weather_year is None:
                msg = (
                    "Weather year can not be None if the model does not provide a year."
                    "Check for correct model name"
                )
                raise ValueError(msg)
            self.study_year = self.config.weather_year

    def build_system(self) -> System:
        """Create infrasys system."""
        logger.info("Building infrasys system using {}", self.__class__.__name__)

        # If we decide to change the engine for handling the data we can do it here.
        object_data = self._plexos_table_data()
        self.plexos_data = self._polarize_data(object_data=object_data)

        # Construct the network
        self._construct_load_zones()
        self._construct_buses()
        self._construct_branches()
        self._construct_interfaces()
        self._construct_reserves()

        # Generators
        self._construct_generators()
        self._add_buses_to_generators()
        self._add_generator_reserves()

        # Batteries
        self._construct_batteries()
        self._add_buses_to_batteries()
        self._add_battery_reserves()

        self._construct_load_profiles()
        return self.system

    def _collect_horizon_data(self, model_name: str) -> dict[str, float]:
        """Collect horizon data (Date From/To) from Plexos database."""
        horizon_query = f"""
        SELECT
            atr.name AS attribute_name,
            COALESCE(attr_data.value, atr.default_value) AS attr_val
        FROM
            t_object
        LEFT JOIN t_class AS class ON
            t_object.class_id == class.class_id
        LEFT JOIN t_attribute AS atr ON
            t_object.class_id  == atr.class_id
        LEFT JOIN t_membership AS tm ON
            t_object.object_id  == tm.child_object_id
        LEFT JOIN t_class AS parent_class ON
            tm.parent_class_id == parent_class.class_id
        LEFT JOIN t_object AS to2 ON
            tm.parent_object_id == to2.object_id
        LEFT JOIN t_attribute_data attr_data ON
            attr_data.attribute_id == atr.attribute_id AND t_object.object_id == attr_data.object_id
        WHERE
            class.name == '{ClassEnum.Horizon.value}'
            AND parent_class.name == '{ClassEnum.Model.value}'
            AND to2.name == '{model_name}'
        """
        horizon_data = self.db.query(horizon_query)
        horizon_map = {key: value for key, value in horizon_data}
        return horizon_map

    def _reconcile_timeseries(self, data_file):
        """
        Adjust timesseries data to match the study year datetime
        index, such as removing or adding leap-year data.
        """
        date_time_column = pd.date_range(
            start=f"1/1/{self.study_year}",
            end=f"1/1/{self.study_year + 1}",
            freq="1h",
            inclusive="left",
        )
        leap_year = len(date_time_column) == 8784

        if data_file.height in [8784, 8760]:
            if data_file.height == 8784 and not leap_year:
                before_feb_29 = data_file.slice(0, 1416)
                after_feb_29 = data_file.slice(1440, len(data_file) - 1440)
                return pl.concat([before_feb_29, after_feb_29])
            elif data_file.height == 8760 and leap_year:
                return handle_leap_year_adjustment(data_file)
            return data_file

        if data_file.height <= 8760:
            return fill_missing_timestamps(data_file, date_time_column)

        if data_file.height in [17568, 17520]:
            return resample_data_to_hourly(data_file)

        return data_file

    def _get_fuel_prices(self):
        logger.debug("Creating fuel representation")
        system_fuels = (pl.col("child_class_name") == ClassEnum.Fuel.value) & (
            pl.col("parent_class_name") == ClassEnum.System.value
        )
        fuels = self._get_model_data(system_fuels)
        fuel_prices = {}
        for fuel_name, fuel_data in fuels.group_by("name"):
            fuel_name = fuel_name[0]
            property_records = fuel_data[
                [
                    "band",
                    "property_name",
                    "property_value",
                    "property_unit",
                    "data_file",
                    "variable",
                    "action",
                    "variable_tag",
                    "timeslice",
                    "timeslice_value",
                ]
            ].to_dicts()

            for property in property_records:
                property.update({"property_unit": "$/MMBtu"})

            mapped_records, multi_band_records = self._parse_property_data(property_records, fuel_name)
            mapped_records["name"] = fuel_name
            fuel_prices[fuel_name] = mapped_records["Price"]
        return fuel_prices

    def _construct_load_zones(self, default_model=LoadZone) -> None:
        """Create LoadZone representation.

        Plexos can define load at multiple levels, but for balancing the load,
        we assume that it happens at the region level, which is a typical way
        of doing it.
        """
        logger.debug("Creating load zone representation")
        system_regions = (pl.col("child_class_name") == ClassEnum.Region.value) & (
            pl.col("parent_class_name") == ClassEnum.System.value
        )
        regions = self._get_model_data(system_regions)

        region_pivot = regions.pivot(  # noqa: PD010
            index=DEFAULT_INDEX,
            columns="property_name",
            values="property_value",
            aggregate_function="first",
        )
        for region in region_pivot.iter_rows(named=True):
            valid_fields, ext_data = field_filter(region, default_model.model_fields)
            valid_fields = prepare_ext_field(valid_fields, ext_data)
            self.system.add_component(default_model(**valid_fields))
        return

    def _construct_buses(self, default_model=ACBus) -> None:
        logger.debug("Creating buses representation")
        system_buses = (pl.col("child_class_name") == ClassEnum.Node.value) & (
            pl.col("parent_class_name") == ClassEnum.System.value
        )
        region_buses = (pl.col("child_class_name") == ClassEnum.Region.value) & (
            pl.col("parent_class_name") == ClassEnum.Node.value
        )
        system_buses = self._get_model_data(system_buses)
        buses_region = self._get_model_data(region_buses)
        buses = system_buses.pivot(  # noqa: PD010
            index=DEFAULT_INDEX,
            columns="property_name",
            values="property_value",
            aggregate_function="first",
        )
        for idx, bus in enumerate(buses.iter_rows(named=True)):
            mapped_bus = {self.property_map.get(key, key): value for key, value in bus.items()}

            valid_fields, ext_data = field_filter(mapped_bus, default_model.model_fields)

            # Get region from buses region memberships
            region_name = buses_region.filter(pl.col("parent_object_id") == bus["object_id"])["name"].item()

            valid_fields["load_zone"] = self.system.get_component(LoadZone, name=region_name)

            # NOTE: We do not parser differenet kind of buses from Plexos
            valid_fields["bus_type"] = ACBusTypes.PV

            valid_fields["base_voltage"] = (
                230.0 if not valid_fields.get("base_voltage") else valid_fields["base_voltage"]
            )

            valid_fields = prepare_ext_field(valid_fields, ext_data)
            self.system.add_component(default_model(number=idx + 1, **valid_fields))
        return

    def _construct_reserves(self, default_model=Reserve):
        logger.debug("Creating reserve representation")
        system_reserves = (pl.col("child_class_name") == ClassEnum.Reserve.name) & (
            pl.col("parent_class_name") == ClassEnum.System.name
        )
        system_reserves = self._get_model_data(system_reserves)

        reserve_pivot = system_reserves.pivot(  # noqa: PD010
            index=DEFAULT_INDEX,
            columns="property_name",
            values="property_value",
            aggregate_function="first",
        )
        for reserve in reserve_pivot.iter_rows(named=True):
            mapped_reserve = {self.property_map.get(key, key): value for key, value in reserve.items()}
            valid_fields, ext_data = field_filter(mapped_reserve, default_model.model_fields)

            if ext_data:
                # Add reserve type and direction based on Plexos type. If the
                # key is not present, the assumed one is the default (Spinning.Up)
                reserve_type = validate_string(ext_data.pop("Type", "default"))

                # Mapping is integer, but parsing reads it as float.
                if isinstance(reserve_type, float):
                    reserve_type = int(reserve_type)

                # If we do not support the reserve type yield the default one.
                plexos_reserve_map = self.config.defaults["reserve_types"].get(
                    str(reserve_type), self.config.defaults["reserve_types"]["default"]
                )  # Pass string so we do not need to convert the json mapping.

                valid_fields["reserve_type"] = ReserveType[plexos_reserve_map["type"]]
                valid_fields["direction"] = ReserveDirection[plexos_reserve_map["direction"]]

                valid_fields = prepare_ext_field(valid_fields, ext_data)

            self.system.add_component(default_model(**valid_fields))

        reserve_map = ReserveMap(name="contributing_generators")
        self.system.add_component(reserve_map)
        return

    def _construct_branches(self, default_model=MonitoredLine):
        logger.debug("Creating lines")
        system_lines = (pl.col("child_class_name") == ClassEnum.Line.value) & (
            pl.col("parent_class_name") == ClassEnum.System.value
        )
        system_lines = self._get_model_data(system_lines)
        lines_pivot = system_lines.pivot(  # noqa: PD010
            index=DEFAULT_INDEX,
            columns="property_name",
            values="property_value",
            aggregate_function="first",
        )

        lines_pivot_memberships = self.db.get_memberships(
            *lines_pivot["name"].to_list(), object_class=ClassEnum.Line
        )
        for line in lines_pivot.iter_rows(named=True):
            line_properties_mapped = {self.property_map.get(key, key): value for key, value in line.items()}
            line_properties_mapped["rating"] = line_properties_mapped.pop("max_power_flow", None)
            line_properties_mapped["rating_up"] = line_properties_mapped.pop("max_power_flow", None)
            line_properties_mapped["rating_down"] = line_properties_mapped.pop("min_power_flow", None)

            valid_fields, ext_data = field_filter(line_properties_mapped, default_model.model_fields)

            from_bus_name = next(
                membership
                for membership in lines_pivot_memberships
                if (
                    membership[2] == line["name"]
                    and membership[5] == ClassEnum.Line.name
                    and membership[6].replace(" ", "") == CollectionEnum.NodeFrom.name
                )
            )[3]
            from_bus = self.system.get_component(ACBus, from_bus_name)
            to_bus_name = next(
                membership
                for membership in lines_pivot_memberships
                if (
                    membership[2] == line["name"]
                    and membership[5] == ClassEnum.Line.name
                    and membership[6].replace(" ", "") == CollectionEnum.NodeTo.name
                )
            )[3]
            to_bus = self.system.get_component(ACBus, to_bus_name)
            valid_fields["from_bus"] = from_bus
            valid_fields["to_bus"] = to_bus

            valid_fields = prepare_ext_field(valid_fields, ext_data)
            self.system.add_component(default_model(**valid_fields))
        return

    def _infer_model_type(self, generator_name):
        inference_mapper = self.device_match_string
        generator_name_lower = generator_name.lower()
        for key, device_info in inference_mapper.items():
            if key in generator_name_lower:
                return device_info
        return None

    def _get_fuel_pmtype(self, generator_name, generator_fuel_map):
        plexos_fuel_name = generator_fuel_map.get(generator_name)
        logger.trace("Parsing generator = {} with fuel type = {}", generator_name, plexos_fuel_name)

        fuel_pmtype = (
            self.device_map.get(generator_name)
            or self.fuel_map.get(plexos_fuel_name)
            or self._infer_model_type(generator_name)
        )

        return fuel_pmtype, plexos_fuel_name

    def _get_model_type(self, fuel_pmtype):
        if fuel_pmtype is not None:
            for model, conditions in self.config.defaults["generator_models"].items():
                for cond in conditions:
                    if (cond["fuel"] == fuel_pmtype["fuel"] or cond["fuel"] is None) and (
                        cond["type"] == fuel_pmtype["type"] or cond["type"] is None
                    ):
                        return model
        return ""

    def _construct_generators(self):
        logger.debug("Creating generators")
        system_generators = (pl.col("child_class_name") == ClassEnum.Generator.name) & (
            pl.col("parent_class_name") == ClassEnum.System.name
        )

        system_generators = self._get_model_data(system_generators)
        if getattr(self.config.feature_flags, "plexos-csv", None):
            system_generators.write_csv("generators.csv")
        # NOTE: The best way to identify the type of generator on Plexos is by reading the fuel
        fuel_query = f"""
        SELECT
            parent_obj.name AS parent_object_name,
            child_obj.name AS fuel_name
        FROM
            t_membership AS mem
            LEFT JOIN t_object AS child_obj ON mem.child_object_id = child_obj.object_id
            LEFT JOIN t_object AS parent_obj ON mem.parent_object_id = parent_obj.object_id
            LEFT JOIN t_class AS child_cls ON child_obj.class_id = child_cls.class_id
            LEFT JOIN t_class AS parent_cls ON parent_obj.class_id = parent_cls.class_id
        WHERE
            child_cls.name = '{ClassEnum.Fuel.value}'
            AND parent_cls.name = '{ClassEnum.Generator.value}'
        """
        generator_fuel = self.db.query(fuel_query)
        generator_fuel_map = {key: value for key, value in generator_fuel}
        fuel_prices = self._get_fuel_prices()

        # Iterate over properties for generator
        for generator_name, generator_data in system_generators.group_by("name"):
            generator_name = generator_name[0]
            fuel_pmtype, plexos_fuel_name = self._get_fuel_pmtype(generator_name, generator_fuel_map)
            model_map = self._get_model_type(fuel_pmtype)
            if getattr(R2X_MODELS, model_map, None) is None:
                logger.warning(
                    "Model map not found for generator={} with fuel_type={}. Skipping it.",
                    generator_name,
                    plexos_fuel_name,
                )
                continue
            fuel_type, pm_type = fuel_pmtype["fuel"], fuel_pmtype["type"]
            model_map = getattr(R2X_MODELS, model_map)

            property_records = generator_data[
                [
                    "band",
                    "property_name",
                    "property_value",
                    "property_unit",
                    "data_file",
                    "variable",
                    "action",
                    "variable_tag",
                    "timeslice",
                    "timeslice_value",
                ]
            ].to_dicts()
            mapped_records, multi_band_records = self._parse_property_data(property_records, generator_name)
            mapped_records["name"] = generator_name

            # if multi_band_records:
            #     pass

            # Add prime mover mapping
            mapped_records["prime_mover_type"] = pm_type or self.config.plexos_fuel_map["default"].get("type")
            mapped_records["prime_mover_type"] = PrimeMoversType[mapped_records["prime_mover_type"]]
            mapped_records["fuel"] = fuel_type or self.config.prime_mover_map["default"].get("fuel")

            # match model_map:
            #     case HydroPumpedStorage():
            #         mapped_records["prime_mover_type"] = PrimeMoversType.PS

            # Pumped Storage generators are not required to have Max Capacity property
            if "max_active_power" not in mapped_records and "pump_load" in mapped_records:
                mapped_records["rating"] = mapped_records["pump_load"]

            mapped_records = self._set_unit_availability(mapped_records)
            if mapped_records is None:
                # When unit availability is not set, we skip the generator
                continue

            required_fields = {
                key: value for key, value in model_map.model_fields.items() if value.is_required()
            }
            if not all(key in mapped_records for key in required_fields):
                missing_fields = [key for key in required_fields if key not in mapped_records]
                logger.warning(
                    "Skipping Generator {}. Missing Required Fields: {}", generator_name, missing_fields
                )
                continue

            mapped_records["fuel_price"] = fuel_prices.get(generator_fuel_map.get(generator_name), 0)
            mapped_records = self._construct_operating_costs(mapped_records, generator_name, model_map)

            mapped_records["base_mva"] = 1

            valid_fields, ext_data = field_filter(mapped_records, model_map.model_fields)

            ts_fields = {k: v for k, v in mapped_records.items() if isinstance(v, SingleTimeSeries)}
            valid_fields.update(
                {
                    ts_name: np.mean(ts.data)
                    for ts_name, ts in ts_fields.items()
                    if ts_name in valid_fields.keys()
                }
            )

            valid_fields = prepare_ext_field(valid_fields, ext_data)
            self.system.add_component(model_map(**valid_fields))
            generator = self.system.get_component_by_label(f"{model_map.__name__}.{generator_name}")

            if ts_fields:
                generator = self.system.get_component_by_label(f"{model_map.__name__}.{generator_name}")
                ts_dict = {"solve_year": self.study_year}
                for ts_name, ts in ts_fields.items():
                    ts.variable_name = ts_name
                    self.system.add_time_series(ts, generator, **ts_dict)

    def _add_buses_to_generators(self):
        # Add buses to generators
        generators = [generator["name"] for generator in self.system.to_records(Generator)]
        generator_memberships = self.db.get_memberships(
            *generators,
            object_class=ClassEnum.Generator,
            collection=CollectionEnum.Nodes,
        )
        for generator in self.system.get_components(Generator):
            buses = [membership for membership in generator_memberships if membership[2] == generator.name]
            if buses:
                for bus in buses:
                    try:
                        bus_object = self.system.get_component(ACBus, name=bus[3])
                    except ISNotStored:
                        logger.warning(
                            "Skipping membership for generator:{} since reserve {} is not stored",
                            generator.name,
                            buses[3],
                        )
                        continue
                    generator.bus = bus_object
        return

    def _add_generator_reserves(self):
        reserve_map = self.system.get_component(ReserveMap, name="contributing_generators")
        generators = [generator["name"] for generator in self.system.to_records(Generator)]
        generator_memberships = self.db.get_memberships(
            *generators,
            object_class=ClassEnum.Generator,
            parent_class=ClassEnum.Reserve,
            collection=CollectionEnum.Generators,
        )
        for generator in self.system.get_components(Generator):
            reserves = [membership for membership in generator_memberships if membership[3] == generator.name]
            if reserves:
                # NOTE: This would get replaced if we have a method on infrasys
                # that check if something exists on the system
                for reserve in reserves:
                    try:
                        reserve_object = self.system.get_component(Reserve, name=reserve[2])
                    except ISNotStored:
                        logger.warning(
                            "Skipping membership for generator:{} since reserve {} is not stored",
                            generator.name,
                            reserve[2],
                        )
                        continue
                    reserve_map.mapping[reserve_object.name].append(generator.name)

        return

    def _construct_batteries(self):
        logger.debug("Creating battery objects")
        system_batteries = self._get_model_data(
            (pl.col("child_class_name") == ClassEnum.Battery.name)
            & (pl.col("parent_class_name") == ClassEnum.System.name)
        )

        required_fields = {
            key: value for key, value in GenericBattery.model_fields.items() if value.is_required()
        }

        for battery_name, battery_data in system_batteries.group_by("name"):
            battery_name = battery_name[0]
            logger.trace("Parsing battery = {}", battery_name)
            property_records = battery_data[
                [
                    "band",
                    "property_name",
                    "property_value",
                    "property_unit",
                    "data_file",
                    "variable",
                    "action",
                    "variable_tag",
                    "timeslice",
                    "timeslice_value",
                ]
            ].to_dicts()

            mapped_records, _ = self._parse_property_data(property_records, battery_name)
            mapped_records["name"] = battery_name

            if "Max Power" in mapped_records:
                mapped_records["rating"] = mapped_records["Max Power"]

            if "Capacity" in mapped_records:
                mapped_records["storage_capacity"] = mapped_records["Capacity"]

            mapped_records["prime_mover_type"] = PrimeMoversType.BA

            valid_fields, ext_data = field_filter(mapped_records, GenericBattery.model_fields)

            valid_fields = self._set_unit_availability(valid_fields)
            if valid_fields is None:
                continue

            if not all(key in valid_fields for key in required_fields):
                missing_fields = [key for key in required_fields if key not in valid_fields]
                logger.warning(
                    "Skipping battery {}. Missing required fields: {}", battery_name, missing_fields
                )
                continue

            if mapped_records["storage_capacity"] == 0:
                logger.warning("Skipping battery {} since it has zero capacity", battery_name)
                continue

            valid_fields = prepare_ext_field(valid_fields, ext_data)
            self.system.add_component(GenericBattery(**valid_fields))
        return

    def _add_buses_to_batteries(self):
        batteries = [battery["name"] for battery in self.system.to_records(GenericBattery)]
        if not batteries:
            msg = "No battery objects found on the system. Skipping adding membership to buses."
            logger.warning(msg)
            return
        generator_memberships = self.db.get_memberships(
            *batteries,
            object_class=ClassEnum.Battery,
            collection=CollectionEnum.Nodes,
        )
        for component in self.system.get_components(GenericBattery):
            buses = [membership for membership in generator_memberships if membership[2] == component.name]
            if buses:
                for bus in buses:
                    try:
                        bus_object = self.system.get_component(ACBus, name=bus[3])
                    except ISNotStored:
                        logger.warning(
                            "Skipping membership for battery:{} since bus {} is not stored",
                            component.name,
                            buses[3],
                        )
                        continue
                    component.bus = bus_object
        return

    def _add_battery_reserves(self):
        reserve_map = self.system.get_component(ReserveMap, name="contributing_generators")
        batteries = [battery["name"] for battery in self.system.to_records(GenericBattery)]
        if not batteries:
            msg = "No battery objects found on the system. Skipping adding membership to buses."
            logger.warning(msg)
            return
        generator_memberships = self.db.get_memberships(
            *batteries,
            object_class=ClassEnum.Battery,
            parent_class=ClassEnum.Reserve,
            collection=CollectionEnum.Batteries,
        )
        for battery in self.system.get_components(GenericBattery):
            reserves = [membership for membership in generator_memberships if membership[3] == battery.name]
            if reserves:
                # NOTE: This would get replaced if we have a method on infrasys
                # that check if something exists on the system
                for reserve in reserves:
                    try:
                        reserve_object = self.system.get_component(Reserve, name=reserve[2])
                    except ISNotStored:
                        logger.warning(
                            "Skipping membership for generator:{} since reserve {} is not stored",
                            battery.name,
                            reserve[2],
                        )
                        continue
                    reserve_map.mapping[reserve_object.name].append(battery.name)
        return

    def _construct_interfaces(self, default_model=TransmissionInterface):
        """Construct Transmission Interface and Transmission Interface Map."""
        logger.debug("Creating transmission interfaces")
        system_interfaces_mask = (pl.col("child_class_name") == ClassEnum.Interface.name) & (
            pl.col("parent_class_name") == ClassEnum.System.name
        )
        system_interfaces = self._get_model_data(system_interfaces_mask)
        interfaces = system_interfaces.pivot(  # noqa: PD010
            index=DEFAULT_INDEX,
            columns="property_name",
            values="property_value",
            aggregate_function="first",
        )

        interface_property_map = {
            v: k
            for k, v in self.config.defaults["plexos_property_map"].items()
            if k in default_model.model_fields
        }

        tx_interface_map = TransmissionInterfaceMap(name="transmission_map")
        for interface in interfaces.iter_rows(named=True):
            mapped_interface = {
                interface_property_map.get(key, key): value for key, value in interface.items()
            }
            valid_fields, ext_data = field_filter(mapped_interface, default_model.model_fields)

            # Check that the interface has all the required fields of the model.
            required_fields = {
                key: value for key, value in default_model.model_fields.items() if value.is_required()
            }
            if not all(key in mapped_interface for key in required_fields):
                missing_fields = [key for key in required_fields if key not in mapped_interface]
                logger.warning(
                    "{}:{} missing required fields: {}. Skipping it.",
                    default_model.__name__,
                    interface["name"],
                    missing_fields,
                )
                continue

            valid_fields = prepare_ext_field(valid_fields, ext_data)
            self.system.add_component(default_model(**valid_fields))

        # Add lines memberships
        lines = [line["name"] for line in self.system.to_records(MonitoredLine)]
        lines_memberships = self.db.get_memberships(
            *lines,
            object_class=ClassEnum.Line,
            parent_class=ClassEnum.Interface,
            collection=CollectionEnum.Lines,
        )
        for line in self.system.get_components(MonitoredLine):
            interface = next(
                (membership for membership in lines_memberships if membership[3] == line.name), None
            )
            if interface:
                # NOTE: This would get replaced if we have a method on infrasys
                # that check if something exists on the system
                try:
                    interface_object = self.system.get_component(TransmissionInterface, name=interface[2])
                except ISNotStored:
                    logger.warning(
                        "Skipping membership for line:{} since interface {} is not stored",
                        line.name,
                        interface[2],
                    )
                    continue
                tx_interface_map.mapping[interface_object.name].append(line.label)
        self.system.add_component(tx_interface_map)
        return

    def _construct_value_curves(self, mapped_records, generator_name):
        """Construct value curves for generators."""
        if any("Heat Rate" in key or "heat_rate" in key for key in mapped_records.keys()):
            vc = None
            heat_rate_avg = mapped_records.get("heat_rate", None)
            heat_rate_base = mapped_records.get("Heat Rate Base", None)
            heat_rate_incr = mapped_records.get("Heat Rate Incr", None)
            heat_rate_incr2 = mapped_records.get("Heat Rate Incr2", None)
            if any(
                isinstance(val, SingleTimeSeries) for val in [heat_rate_avg, heat_rate_base, heat_rate_incr]
            ):
                logger.warning("Market-Bid Cost not implemented for generator={}", generator_name)
                return mapped_records
            if heat_rate_avg:
                fn = LinearFunctionData(proportional_term=heat_rate_avg.magnitude, constant_term=0)
                vc = AverageRateCurve(
                    name=f"{generator_name}_HR",
                    function_data=fn,
                    initial_input=heat_rate_avg.magnitude,
                )
            elif heat_rate_incr2 and "** 2" in str(heat_rate_incr2.units):
                fn = QuadraticFunctionData(
                    quadratic_term=heat_rate_incr2.magnitude,
                    proportional_term=heat_rate_incr.magnitude,
                    constant_term=heat_rate_base.magnitude,
                )
            elif not heat_rate_incr2 and heat_rate_incr:
                fn = LinearFunctionData(
                    proportional_term=heat_rate_incr.magnitude, constant_term=heat_rate_base.magnitude
                )
            else:
                logger.warning("Heat Rate type not implemented for generator={}", generator_name)
                fn = None
            if not vc:
                vc = InputOutputCurve(name=f"{generator_name}_HR", function_data=fn)
            mapped_records["hr_value_curve"] = vc
        return mapped_records

    def _construct_operating_costs(self, mapped_records, generator_name, model_map):
        """Construct operating costs from Value Curves and Operating Costs."""
        mapped_records = self._construct_value_curves(mapped_records, generator_name)
        hr_curve = mapped_records.get("hr_value_curve")

        if issubclass(model_map, RenewableGen):
            mapped_records["operation_cost"] = RenewableGenerationCost()
        elif issubclass(model_map, ThermalGen):
            if hr_curve:
                fuel_cost = mapped_records["fuel_price"]
                if isinstance(fuel_cost, SingleTimeSeries):
                    fuel_cost = np.mean(fuel_cost.data)
                elif isinstance(fuel_cost, Quantity):
                    fuel_cost = fuel_cost.magnitude
                fuel_curve = FuelCurve(value_curve=hr_curve, fuel_cost=fuel_cost)
                mapped_records["operation_cost"] = ThermalGenerationCost(variable=fuel_curve)
                mapped_records.pop("hr_value_curve")
            else:
                logger.warning("No heat rate curve found for generator={}", generator_name)
        elif issubclass(model_map, HydroDispatch):
            mapped_records["operation_cost"] = HydroGenerationCost()
        else:
            logger.warning(
                "Operating Cost not implemented for generator={} model map={}", generator_name, model_map
            )
        return mapped_records

    def _select_model_name(self):
        # TODO(pesap): Handle exception if no model name found
        # https://github.com/NREL/R2X/issues/10
        query = f"""
        select obj.name
        from t_object as obj
        left join t_class as cls on obj.class_id = cls.class_id
        where cls.name = '{ClassEnum.Model.name}'
        """
        result = self.db.query(query)
        model_names = [row[0] for row in result]

        print("Available models:")
        for idx, name in enumerate(model_names):
            print(f"{idx + 1}. {name}")

        while True:
            try:
                choice = int(input("Select a model by number: "))
                if 1 <= choice <= len(model_names):
                    return model_names[choice - 1]
                else:
                    print(f"Please enter a number between 1 and {len(model_names)}.")
            except ValueError:
                print("Invalid input. Please enter a number.")

    def _process_scenarios(self, model_name: str | None = None) -> None:
        """Create a SQLite representation of the XML."""
        if model_name is None:
            msg = "Required model name not found. Parser requires a model to parse from the Plexos database"
            raise ModelError(msg)
        # self.db = PlexosSQLite(xml_fname=xml_fpath)

        logger.trace("Getting object_id for model={}", model_name)
        model_id = self.db.query("select object_id from t_object where name = ?", params=(model_name,))
        if len(model_id) > 1:
            msg = f"Multiple models with the same {model_name} returned. Check database or spelling."
            logger.debug(model_id)
            raise ModelError(msg)
        if not model_id:
            msg = f"Model `{model_name}` not found on the XML. Check spelling of the `model_name`."
            raise ParserError(msg)
        self.model_id = model_id[0][0]  # Unpacking tuple [(model_id,)]

        # NOTE: When doing performance updates this query could get some love.
        valid_scenarios = self.db.query(
            "select obj.name from t_membership mem "
            "left join t_object as obj on obj.object_id = mem.child_object_id "
            "left join t_class as cls on cls.class_id = obj.class_id "
            f"where mem.parent_object_id = {self.model_id} and cls.name = '{ClassEnum.Scenario}'"
        )
        if not valid_scenarios:
            msg = f"{model_name=} does not have any scenario attached to it."
            logger.warning(msg)
            return
        assert valid_scenarios
        self.scenarios = [scenario[0] for scenario in valid_scenarios]  # Flatten list of tuples
        return None

    def _set_unit_availability(self, mapped_records):
        """
        Set availability and active power limit TS for generators.
        Note: variables use infrasys naming scheme, rating != plexos rating.
        """
        # TODO @ktehranchi: #35 Include date_from and date_to in the availability
        # https://github.com/NREL/R2X/issues/35

        availability = mapped_records.get("available", None)
        if availability is not None and availability > 0:
            # Set availability, rating, storage_capacity as multiplier of availability/'units'
            if mapped_records.get("storage_capacity") is not None:
                mapped_records["storage_capacity"] *= mapped_records.get("available")
            mapped_records["rating"] = mapped_records.get("rating") * mapped_records.get("available")
            mapped_records["available"] = 1

            # Set active power limits
            rating_factor = mapped_records.get("Rating Factor", 100)
            rating_factor = self._apply_action(np.divide, rating_factor, 100)
            rating = mapped_records.get("rating", None)
            max_active_power = mapped_records.get("max_active_power", None)
            min_energy_hour = mapped_records.get("Min Energy Hour", None)

            if isinstance(max_active_power, dict):
                max_active_power = self._time_slice_handler("max_active_power", max_active_power)

            if max_active_power is not None:
                units = max_active_power.units
                val = self._apply_action(np.multiply, rating_factor, max_active_power)
            elif rating is not None:
                units = rating.units
                val = self._apply_action(np.multiply, rating_factor, rating.magnitude)
            else:
                return mapped_records
            val = self._apply_unit(val, units)

            if min_energy_hour is not None:
                mapped_records["min_active_power"] = mapped_records.pop("Min Energy Hour")

            if isinstance(val, SingleTimeSeries):
                val.variable_name = "max_active_power"
                self._apply_action(np.divide, val, rating.magnitude)
                mapped_records["max_active_power"] = val
            mapped_records["active_power"] = rating

            if isinstance(rating_factor, SingleTimeSeries):
                mapped_records.pop("Rating Factor")  # rm for ext exporter

        else:  # if unit field not activated in model, skip generator
            mapped_records = None
        return mapped_records

    def _plexos_table_data(self) -> list[tuple]:
        # Get objects table/membership table
        sql_query = files("plexosdb.queries").joinpath("object_query.sql").read_text()
        object_data = self.db.query(sql_query)
        return object_data

    def _polarize_data(self, object_data: list[tuple]) -> pl.DataFrame:
        return pl.from_records(object_data, schema=DEFAULT_QUERY_COLUMNS_SCHEMA)

    def _get_model_data(self, data_filter) -> pl.DataFrame:
        """Filter plexos data for a given class and all scenarios in a model."""
        scenario_specific_data = None
        scenario_filter = None
        if getattr(self, "scenarios", None):
            scenario_filter = pl.col("scenario").is_in(self.scenarios)
            scenario_specific_data = self.plexos_data.filter(data_filter & scenario_filter)
            scenario_specific_data = filter_property_dates(scenario_specific_data, self.study_year)

        base_case_filter = pl.col("scenario").is_null()
        # Default is to parse data normally if there is not scenario. If scenario exist modify the filter.
        if scenario_specific_data is None:
            system_data = self.plexos_data.filter(data_filter & base_case_filter)
            system_data = filter_property_dates(system_data, self.study_year)
        else:
            # include both scenario specific and basecase data
            combined_key_base = pl.col("name") + "_" + pl.col("property_name")
            combined_key_scenario = (
                scenario_specific_data["name"] + "_" + scenario_specific_data["property_name"]
            )

            base_case_filter = base_case_filter & (
                ~combined_key_base.is_in(combined_key_scenario) | pl.col("property_name").is_null()
            )
            base_case_data = self.plexos_data.filter(data_filter & base_case_filter)
            base_case_data = filter_property_dates(base_case_data, self.study_year)

            system_data = pl.concat([scenario_specific_data, base_case_data])

        # If date_from / date_to is specified, override the base_case_value
        rows_to_keep = system_data.filter(pl.col("date_from").is_not_null() | pl.col("date_to").is_not_null())
        rtk_key = rows_to_keep["name"] + "_" + rows_to_keep["property_name"]
        sys_data_key = system_data["name"] + "_" + system_data["property_name"]
        if not rows_to_keep.is_empty():
            # remove if name and property_name are the same
            system_data = system_data.filter(~sys_data_key.is_in(rtk_key))
            system_data = pl.concat([system_data, rows_to_keep])

        # Get System Variables
        variable_data = None
        variable_filter = (
            (pl.col("child_class_name") == ClassEnum.Variable.name)
            & (pl.col("parent_class_name") == ClassEnum.System.name)
            & (pl.col("data_file").is_not_null())
        )
        variable_scenario_data = None
        if scenario_specific_data is not None and scenario_filter is not None:
            variable_scenario_data = self.plexos_data.filter(variable_filter & scenario_filter)

        if variable_scenario_data is not None:
            variable_base_data = self.plexos_data.filter(variable_filter & pl.col("scenario").is_null())
        else:
            variable_base_data = self.plexos_data.filter(variable_filter & base_case_filter)
        if variable_base_data is not None and variable_scenario_data is not None:
            variable_data = pl.concat([variable_scenario_data, variable_base_data])

        return self._join_variable_data(system_data, variable_data)

    def _join_variable_data(self, system_data, variable_data):
        """Join system data with variable data."""
        # Filter Variables
        if variable_data is not None:
            results = []
            grouped = variable_data.group_by("name")
            for group_name, group_df in grouped:
                if group_df.height > 1:
                    # Check if any scenario_name exists
                    scenario_exists = group_df.filter(pl.col("scenario").is_not_null())

                    if scenario_exists.height > 0:
                        # Select the first row with a scenario_name
                        selected_row = scenario_exists[0]
                    else:
                        # If no scenario_name, select the row with the lowest band_id
                        selected_row = group_df.sort("band").head(1)[0]
                else:
                    # If the group has only one row, select that row
                    selected_row = group_df[0]

                results.append(
                    {
                        "name": group_name[0],
                        "variable_name": selected_row["data_file_tag"][0],
                        "variable": selected_row["data_file"][0],
                    }
                )
            variables_filtered = pl.DataFrame(results)
            system_data = system_data.join(
                variables_filtered, left_on="variable_tag", right_on="name", how="left"
            )
        else:
            # NOTE: We might want to include this at the instead of each function call
            system_data = system_data.with_columns(
                pl.lit(None).alias("variable"), pl.lit(None).alias("variable_name")
            )
        system_data = system_data.join(
            variables_filtered, left_on="variable_tag", right_on="name", how="left"
        )
        return system_data

    def _construct_load_profiles(self):
        logger.debug("Creating load profile representation")
        system_regions = (pl.col("child_class_name") == ClassEnum.Region.name) & (
            pl.col("parent_class_name") == ClassEnum.System.name
        )
        regions = self._get_model_data(system_regions).filter(~pl.col("variable").is_null())
        for region, region_data in regions.group_by("name"):
            if not len(region_data) == 1:
                msg = (
                    "load data has more than one row for {}. Selecting the first match. "
                    "Check filtering of properties"
                )
                logger.warning(msg, region)

            ts = self._csv_file_handler(
                property_name="max_active_power", property_data=region_data["variable"][0]
            )
            max_load = np.max(ts.data.to_numpy())
            ts = self._apply_action(np.divide, ts, max_load)
            if not ts:
                continue

            bus_region_membership = self.db.get_memberships(
                region[0],
                object_class=ClassEnum.Region,
                parent_class=ClassEnum.Node,
                collection=CollectionEnum.Region,
            )
            for bus in bus_region_membership:
                bus = self.system.get_component(ACBus, name=bus[2])
                load = PowerLoad(
                    name=f"{bus.name}",
                    bus=bus,
                    max_active_power=float(max_load / len(bus_region_membership)) * ureg.MW,
                )
                self.system.add_component(load)
                ts_dict = {"solve_year": self.study_year}

                self.system.add_time_series(ts, load, **ts_dict)
        return

    def _csv_file_handler(self, property_name, property_data):
        fpath_text = property_data
        if "\\" in fpath_text:
            relative_path = PureWindowsPath(fpath_text)
        else:
            relative_path = Path(fpath_text)
        assert relative_path
        assert self.config.run_folder
        fpath = self.config.run_folder / relative_path
        try:
            data_file = pl.read_csv(
                fpath.as_posix(), infer_schema_length=10000
            )  # This might not work on Windows machines
        except FileNotFoundError:
            logger.warning("File {} not found. Skipping it.", relative_path)
            return

        # Lowercase files
        data_file = data_file.with_columns(pl.col(pl.String).str.to_lowercase()).rename(
            {column: column.lower() for column in data_file.columns}
        )

        single_value_data = self._retrieve_single_value_data(property_name, data_file)
        if single_value_data is not None:
            return single_value_data

        time_series_data = self._retrieve_time_series_data(property_name, data_file)
        if time_series_data is not None:
            return time_series_data
        logger.debug("Skipped file {}", relative_path)

    def _retrieve_single_value_data(self, property_name, data_file):
        if (
            all(column in data_file.columns for column in [property_name.lower(), "year"])
            and "month" not in data_file.columns
        ):
            data_file = data_file.filter(pl.col("year") == self.study_year)
            return data_file[property_name.lower()][0]

        if data_file.columns[:2] == PROPERTY_SV_COLUMNS_BASIC:
            return data_file.filter(pl.col("name") == property_name.lower())["value"][0]

        if data_file.columns == PROPERTY_SV_COLUMNS_NAMEYEAR:  # double check this case
            filter_condition = (pl.col("year") == self.study_year) & (pl.col("name") == property_name.lower())
            try:
                return data_file.filter(filter_condition)["value"][0]
            except IndexError:
                logger.warning("Property {} missing data_file data. Skipping it.", property_name)
                return
        return

    def _retrieve_time_series_data(self, property_name, data_file):
        output_columns = ["year", "month", "day", "hour", "value"]

        if all(column in data_file.columns for column in PROPERTY_TS_COLUMNS_MONTH_PIVOT):
            # Convert these types to ["pattern", "value"]
            data_file = data_file.filter(pl.col("name") == property_name.lower())
            data_file = data_file.melt(id_vars=["name"], variable_name="pattern")
            data_file = data_file.select(["pattern", "value"])

        match data_file.columns:
            case ["pattern", "value"]:
                dt = pl.datetime_range(
                    datetime(self.study_year, 1, 1), datetime(self.study_year, 12, 31), "1h", eager=True
                ).alias("datetime")
                date_df = pl.DataFrame({"datetime": dt})
                date_df = date_df.with_columns(
                    [
                        date_df["datetime"].dt.year().alias("year"),
                        date_df["datetime"].dt.month().alias("month"),
                        date_df["datetime"].dt.day().alias("day"),
                        date_df["datetime"].dt.hour().alias("hour"),
                    ]
                )

                data_file = data_file.with_columns(
                    month=pl.col("pattern").str.extract(r"(\d{2})$").cast(pl.Int8)
                )  # If other patterns exist, this will need to change.
                data_file = date_df.join(data_file.select("month", "value"), on="month", how="inner").select(
                    output_columns
                )

            case ["month", "day", "period", "value"]:
                data_file = data_file.rename({"period": "hour"})
                data_file = data_file.with_columns(pl.lit(self.study_year).alias("year"))
                columns = ["year"] + [col for col in data_file.columns if col != "year"]
                data_file = data_file.select(columns)

            case columns if all(
                column in columns for column in [*PROPERTY_TS_COLUMNS_MDP, property_name.lower()]
            ):
                data_file = data_file.with_columns(pl.lit(self.study_year).alias("year"))
                data_file = data_file.rename({property_name.lower(): "value"})
                data_file = data_file.rename({"period": "hour"})
                data_file = data_file.select(output_columns)

            case columns if all(
                column in columns for column in [*PROPERTY_TS_COLUMNS_YM, property_name.lower()]
            ):
                data_file = data_file.filter(pl.col("year") == self.study_year)
                data_file = data_file.rename({property_name.lower(): "value"})
                data_file = data_file.with_columns(day=pl.lit(1), hour=pl.lit(0))
                data_file = data_file.select(output_columns)

            case columns if all(column in columns for column in PROPERTY_TS_COLUMNS_BASIC):
                data_file = data_file.rename({"period": "hour"})
                data_file = data_file.filter(pl.col("year") == self.study_year)

            # case columns if all(column in columns for column in PROPERTY_TS_COLUMNS_MULTIZONE):
            #     # Need to test these file types still
            #     # Drop all columns that are not datetime columns or the region name
            #     data_file = data_file.filter(pl.col("year") == self.study_year)
            #     data_file = data_file.drop(
            #         *[col for col in data_file.columns if col not in DATETIME_COLUMNS_MULTIZONE + [region]]
            #     )
            #     # Rename region name to hour
            #     data_file = data_file.rename({region: "value"})

            case columns if all(column in columns for column in PROPERTY_TS_COLUMNS_PIVOT):
                data_file = data_file.filter(pl.col("year") == self.study_year)
                data_file = data_file.melt(id_vars=PROPERTY_TS_COLUMNS_PIVOT, variable_name="hour")
                data_file = data_file.with_columns(pl.col("hour").cast(pl.Int8))

            case _:
                logger.warning("Data file columns not supported. Skipping it.")
                logger.warning("Datafile Columns: {}", data_file.columns)
                return

        if data_file.is_empty():
            logger.debug("Weather year doesn't exist in {}. Skipping it.", property_name)
            return

        # Format to SingleTimeSeries
        if data_file.columns == output_columns:
            data_file = data_file.sort(["year", "month", "day", "hour"])
            resolution = timedelta(hours=1)
            first_row = data_file.row(0)
            start = datetime(year=first_row[0], month=first_row[1], day=first_row[2])
            variable_name = property_name  # Change with property mapping
            data_file = self._reconcile_timeseries(data_file)
            return SingleTimeSeries.from_array(
                data=data_file["value"].cast(pl.Float64),
                resolution=resolution,
                initial_time=start,
                variable_name=variable_name,
                # Maybe change this to be the property name rather than the object name?
            )
        return

    def _time_slice_handler(self, property_name, property_data):
        """Deconstructs dict of timeslices into SingleTimeSeries objects."""
        resolution = timedelta(hours=1)
        initial_time = datetime(self.study_year, 1, 1)
        date_time_array = np.arange(
            f"{self.study_year}",
            f"{self.study_year + 1}",
            dtype="datetime64[h]",
        )  # Removing 1 day to match ReEDS convention and converting into a vector
        months = np.array([dt.astype("datetime64[M]").astype(int) % 12 + 1 for dt in date_time_array])
        month_datetime_series = np.zeros(len(date_time_array), dtype=float)

        # Helper function to parse the key patterns
        def parse_key(key):
            # Split by semicolons for multiple ranges
            ranges = key.split(";")
            month_list = []
            for rng in ranges:
                # Match ranges like 'M5-10' and single months like 'M1'
                match = re.match(r"M(\d+)(?:-(\d+))?", rng)
                if match:
                    start_month = int(match.group(1))
                    end_month = int(match.group(2)) if match.group(2) else start_month
                    # Generate the list of months from the range
                    month_list.extend(range(start_month, end_month + 1))
            return month_list

        # Fill the month_datetime_series with the property data values
        for key, value in property_data.items():
            months_in_key = parse_key(key)
            # Set the value in the array for the corresponding months
            for month in months_in_key:
                month_datetime_series[months == month] = value.magnitude

        return SingleTimeSeries.from_array(
            month_datetime_series,
            property_name,
            initial_time=initial_time,
            resolution=resolution,
        )

    def _apply_unit(self, value, unit):
        if isinstance(value, SingleTimeSeries):
            value.units = str(unit)
            return value
        return value * unit if unit else value

    def _apply_action(self, action, val_a, val_b):
        val_a_data = val_a.data if isinstance(val_a, SingleTimeSeries) else val_a
        val_b_data = val_b.data if isinstance(val_b, SingleTimeSeries) else val_b

        results = action(val_a_data, val_b_data)
        if isinstance(val_a, SingleTimeSeries):
            val_a.data = results
            return val_a
        if isinstance(val_b, SingleTimeSeries):
            val_b.data = results
            return val_b
        return results

    def _get_value(self, prop_value, unit, record, record_name):
        """Parse Property value from record csv, timeslice, and datafiles."""
        data_file = (
            self._csv_file_handler(record_name, record.get("data_file")) if record.get("data_file") else None
        )
        if data_file is None and record.get("data_file"):
            return None

        variable = (
            self._csv_file_handler(record.get("variable_tag"), record.get("variable"))
            if record.get("variable")
            else None
        )

        actions = {
            "×": np.multiply,  # noqa
            "+": np.add,
            "-": np.subtract,
            "/": np.divide,
            "=": lambda x, y: y,
        }
        action = actions[record.get("action")] if record.get("action") else None
        timeslice_value = record.get("timeslice_value")

        if variable is not None:
            if record.get("action") == "=":
                return self._apply_unit(variable, unit)
            if data_file is not None:
                return self._apply_unit(self._apply_action(action, variable, data_file), unit)
            if prop_value is not None:
                return self._apply_unit(self._apply_action(action, variable, prop_value), unit)
            return self._apply_unit(variable, unit)

        if data_file is not None:
            if timeslice_value is not None and timeslice_value != -1:
                return self._apply_unit(self._apply_action(action, data_file, timeslice_value), unit)
            return self._apply_unit(data_file, unit)

        if timeslice_value is not None and timeslice_value != -1:
            return self._apply_unit(timeslice_value, unit)

        return self._apply_unit(prop_value, unit)

    def _parse_property_data(self, record_data, record_name):
        mapped_properties = {}
        property_counts = {}
        multi_band_properties = set()

        for record in record_data:
            band = record["band"]
            timeslice = record["timeslice"]
            prop_name = record["property_name"]
            prop_value = record["property_value"]
            unit = record["property_unit"].replace("$", "usd")

            mapped_property_name = self.property_map.get(prop_name, prop_name)
            unit = None if unit == "-" else unit
            if unit:
                try:
                    unit = ureg[unit]
                except UndefinedUnitError:
                    unit = None

            value = self._get_value(prop_value, unit, record, record_name)
            if value is None:
                logger.warning("Property {} missing record data for {}. Skipping it.", prop_name, record_name)
                continue

            if timeslice is not None:
                # Timeslice Properties
                if mapped_property_name not in property_counts:
                    # First Time reading timeslice
                    nested_dict = {}
                    nested_dict[timeslice] = value
                    mapped_properties[mapped_property_name] = nested_dict
                    property_counts[mapped_property_name] = {timeslice}
                elif timeslice not in property_counts[mapped_property_name]:
                    mapped_properties[mapped_property_name][timeslice] = value
                    property_counts[mapped_property_name].add(timeslice)
                    multi_band_properties.add(mapped_property_name)
            else:
                # Standard Properties
                if mapped_property_name not in property_counts:
                    # First time reading basic property
                    mapped_properties[mapped_property_name] = value
                    property_counts[mapped_property_name] = {band}
                else:
                    # Multi-band properties
                    if band not in property_counts[mapped_property_name]:
                        new_prop_name = f"{mapped_property_name}_{band}"
                        mapped_properties[new_prop_name] = value
                        property_counts[mapped_property_name].add(band)
                        multi_band_properties.add(mapped_property_name)
                    else:
                        logger.warning(
                            "Property {} for {} has multiple values specified. Using the last one.",
                            mapped_property_name,
                            record_name,
                        )
                        mapped_properties[mapped_property_name] = value
        return mapped_properties, multi_band_properties


if __name__ == "__main__":
    from ..logger import setup_logging
    from .handler import get_parser_data

    run_folder = Path("")
    # Functions relative to the parser.
    setup_logging(level="DEBUG")

    config = Scenario.from_kwargs(
        name="Plexos-Test",
        input_model="plexos",
        run_folder=run_folder,
        solve_year=2030,
        weather_year=2030,
        model="",
    )
    config.fmap["xml_file"]["fname"] = ""

    parser = get_parser_data(config=config, parser_class=PlexosParser)



"""Cost related functions."""

from infrasys import Component
from typing import Annotated
from pydantic import Field
from infrasys.cost_curves import ProductionVariableCostCurve
from r2x.units import FuelPrice


class OperationalCost(Component):
    name: Annotated[str, Field(frozen=True)] = ""


class RenewableGenerationCost(OperationalCost):
    curtailment_cost: ProductionVariableCostCurve | None = None
    variable: ProductionVariableCostCurve | None = None


class HydroGenerationCost(OperationalCost):
    fixed: Annotated[FuelPrice, Field(description="Cost of using fuel in $/MWh.")] = FuelPrice(0.0, "usd/MWh")
    variable: ProductionVariableCostCurve | None = None


class ThermalGenerationCost(OperationalCost):
    start_up: Annotated[FuelPrice, Field(description="Cost of using fuel in $/MWh.")] = FuelPrice(
        0.0, "usd/MWh"
    )
    fixed: Annotated[FuelPrice, Field(description="Cost of using fuel in $/MWh.")] = FuelPrice(0.0, "usd/MWh")
    shut_down: Annotated[FuelPrice, Field(description="Cost of using fuel in $/MWh.")] = FuelPrice(
        0.0, "usd/MWh"
    )
    variable: ProductionVariableCostCurve | None = None


class StorageCost(OperationalCost):
    start_up: Annotated[FuelPrice, Field(description="Cost of using fuel in $/MWh.")] = FuelPrice(
        0.0, "usd/MWh"
    )
    fixed: Annotated[FuelPrice, Field(description="Cost of using fuel in $/MWh.")] = FuelPrice(0.0, "usd/MWh")
    shut_down: Annotated[FuelPrice, Field(description="Cost of using fuel in $/MWh.")] = FuelPrice(
        0.0, "usd/MWh"
    )
    energy_surplus_cost: Annotated[FuelPrice, Field(description="Cost of using fuel in $/MWh.")] = FuelPrice(
        0.0, "usd/MWh"
    )
    energy_storage_cost: Annotated[FuelPrice, Field(description="Cost of using fuel in $/MWh.")] = FuelPrice(
        0.0, "usd/MWh"
    )
    charge_variable_cost: ProductionVariableCostCurve | None = None
    discharge_variable_cost: ProductionVariableCostCurve | None = None



"""Useful function for models."""

from .generators import Generator, ThermalGen, HydroGen, Storage, RenewableGen
from .costs import (
    OperationalCost,
    ThermalGenerationCost,
    StorageCost,
    HydroGenerationCost,
    RenewableGenerationCost,
)
from loguru import logger


def get_operational_cost(model: type["Generator"]) -> type["OperationalCost"] | None:
    """Return operational cost for the type of generator model."""
    match model:
        case _ if issubclass(model, ThermalGen):
            return ThermalGenerationCost
        case _ if issubclass(model, HydroGen):
            return HydroGenerationCost
        case _ if issubclass(model, RenewableGen):
            return RenewableGenerationCost
        case _ if issubclass(model, Storage):
            return StorageCost
        case _:
            msg = (
                f"{model=} does not have an operational cost. "
                "Check that a operational cost exist for the model."
            )
            logger.warning(msg)
            return None



from plexosdb.sqlite import PlexosSQLite
import pytest
from plexosdb import XMLHandler
from r2x.api import System
from r2x.config import Scenario
from r2x.exceptions import ParserError
from r2x.parser.handler import get_parser_data
from r2x.parser.plexos import PlexosParser

DB_NAME = "2-bus_example.xml"
MODEL_NAME = "main_model"


@pytest.fixture
def plexos_scenario(tmp_path, data_folder):
    return Scenario.from_kwargs(
        name="plexos_test",
        input_model="plexos",
        run_folder=data_folder,
        output_folder=tmp_path,
        model=MODEL_NAME,
        solve_year=2035,
        weather_year=2012,
        fmap={"xml_file": {"fname": DB_NAME, "model": "default"}},
    )


@pytest.fixture
def plexos_parser_instance(plexos_scenario):
    plexos_device_map = {"SolarPV_01": "RenewableFix", "ThermalCC": "ThermalStandard"}
    plexos_scenario.defaults["plexos_device_map"] = plexos_device_map
    return get_parser_data(plexos_scenario, parser_class=PlexosParser)


def test_plexos_parser_instance(plexos_parser_instance):
    assert isinstance(plexos_parser_instance, PlexosParser)
    assert len(plexos_parser_instance.data) == 1  # Plexos parser just parses a single file
    assert isinstance(plexos_parser_instance.data["xml_file"], XMLHandler)
    assert isinstance(plexos_parser_instance.db, PlexosSQLite)


@pytest.mark.skip(reason="We need a better test model")
def test_build_system(plexos_parser_instance):
    system = plexos_parser_instance.build_system()
    assert isinstance(system, System)


def test_raise_if_no_map_provided(tmp_path, data_folder):
    scenario = Scenario.from_kwargs(
        name="plexos_test",
        input_model="plexos",
        run_folder=data_folder,
        output_folder=tmp_path,
        solve_year=2035,
        model=MODEL_NAME,
        weather_year=2012,
        fmap={"xml_file": {"fname": DB_NAME}},
    )
    with pytest.raises(ParserError):
        _ = get_parser_data(scenario, parser_class=PlexosParser)



import pytest
from r2x.__version__ import __data_model_version__
from r2x.api import System
from r2x.models import Area, ACBus, Generator
from r2x.units import ureg


@pytest.fixture(scope="class")
def empty_system():
    return System(name="TestSystem")


def test_system_instance(empty_system):
    assert isinstance(empty_system, System)


def test_system_data_model_version(empty_system):
    assert empty_system.version == __data_model_version__
    assert empty_system.data_format_version == __data_model_version__


def test_add_single_component(empty_system):
    area = Area.example()
    empty_system.add_component(area)
    assert isinstance(empty_system.get_component(Area, area.name), Area)


def test_add_composed_component():
    system = System(name="TestComposed", auto_add_composed_components=True)

    # Simple scenario of Generator with a bus attached
    bus = ACBus.example()
    generator = Generator(name="TestGen", active_power=100 * ureg.MW, bus=bus)
    system.add_component(generator)

    assert system.get_component(Generator, "TestGen") == generator
    assert system.get_component(Generator, "TestGen").bus == bus



"""R2X pint units."""

from infrasys.base_quantity import ureg, BaseQuantity

# ruff: noqa
# type: ignore

ureg.define("MMBtu = [energy] = 293071 * [watt_hour]")
ureg.define("usd = []")


class Distance(BaseQuantity):
    __base_unit__ = "meter"


class Voltage(BaseQuantity):
    __base_unit__ = "kilovolt"


class Current(BaseQuantity):
    __base_unit__ = "ampere"


class Angle(BaseQuantity):
    __base_unit__ = "degree"


class ActivePower(BaseQuantity):
    __base_unit__ = "megawatt"


class ApparentPower(BaseQuantity):
    __base_unit__ = "volt_ampere"


class Time(BaseQuantity):
    __base_unit__ = "minute"


class Resistance(BaseQuantity):
    __base_unit__ = "ohm"


class HeatRate(BaseQuantity):
    __base_unit__ = "Btu/kWh"


class FuelPrice(BaseQuantity):
    __base_unit__ = "usd/watthour"


class Energy(BaseQuantity):
    __base_unit__ = "watthour"


class Percentage(BaseQuantity):
    __base_unit__ = "percent"


class EmissionRate(BaseQuantity):
    __base_unit__ = "kg/MWh"


class PowerRate(BaseQuantity):
    __base_unit__ = "MW/min"


def get_magnitude(field) -> float | int:
    """Get reference base power of the component."""
    return field.magnitude if isinstance(field, BaseQuantity) else field



"""Definition of enums used on R2X.

In R2X most of the enums are `StrEnums` to be able to export it to csv easier.
"""

from enum import StrEnum


class ReserveType(StrEnum):
    """Class representing different types of Reserves."""

    SPINNING = "SPINNING"
    FLEXIBILITY = "FLEXIBILITY"
    REGULATION = "REGULATION"


class ReserveDirection(StrEnum):
    """Class representing different Reserve Direction."""

    UP = "UP"
    DOWN = "DOWN"


class ACBusTypes(StrEnum):
    """Enum to define quantities for load flow calculation and categorize buses.

    For PCM translations, must of the buses are `PV`.
    """

    PV = "PV"
    PQ = "PQ"
    REF = "REF"


class PrimeMoversType(StrEnum):
    """EIA prime mover codes."""

    BA = "BA"
    BT = "BT"
    CA = "CA"
    CC = "CC"
    CE = "CE"
    CP = "CP"
    CS = "CSV"
    CT = "CT"
    ES = "ES"
    FC = "FC"
    FW = "FW"
    GT = "GT"
    HA = "HA"
    HB = "HB"
    HK = "HK"
    HY = "HY"
    IC = "IC"
    PS = "PS"
    OT = "OT"
    ST = "ST"
    PV = "PV"
    WT = "WT"
    WS = "WS"
    RTPV = "RTPV"


class ThermalFuels(StrEnum):
    """Thermal fuels that reflect options in the EIA annual energy review."""

    COAL = "COAL"
    WASTE_COAL = "WASTE_COAL"
    DISTILLATE_FUEL_OIL = "DISTILLATE_FUEL_OIL"
    WASTE_OIL = "WASTE_OIL"
    PETROLEUM_COKE = "PETROLEUM_COKE"
    RESIDUAL_FUEL_OIL = "RESIDUAL_FUEL_OIL"
    NATURAL_GAS = "NATURAL_GAS"
    OTHER_GAS = "OTHER_GAS"
    NUCLEAR = "NUCLEAR"
    AG_BIOPRODUCT = "AG_BIOPRODUCT"
    MUNICIPAL_WASTE = "MUNICIPAL_WASTE"
    WOOD_WASTE = "WOOD_WASTE"
    GEOTHERMAL = "GEOTHERMAL"
    OTHER = "OTHER"



"""Helper function to validate data."""

# System packages
import os
from collections.abc import Iterable
from typing import Any

# Third-party packages
import pandas as pd
from loguru import logger

# Module packages
from .utils import DEFAULT_COLUMN_MAP


def check_input_files(run_folder: str, fmap: dict[str, Any]):
    """Validate ReEDS input folder.

    This function wraps some of the test we apply to validate the input files
    from the ReEDS model.

    Args:
        run_folder: Folder to look for the files,
        fmap: File project mapping.

    TODO: We can probably create this function to take a callable function that
    could be used to test inputs from other models as well.
    """
    # Get a list of mandatory files.
    file_list = [value["fname"] for _, value in fmap.items() if value.get("mandatory", False)]
    file_column_dict = {
        value["fname"]: value.get("column_mapping").keys()
        for _, value in fmap.items()
        if value.get("column_mapping")  # Only process files with column_mapping
        if value.get("mandatory", False)
    }

    # Verify files first.
    missing_files = get_missing_files(run_folder, file_list)
    assert len(missing_files) == 0, f"The following files are missing: {missing_files}"

    # Verify columns needed.

    missing_columns = []
    for fname, expected_columns in file_column_dict.items():
        for path_prefix in ["outputs", "inputs_case", "outputs/inputs_params"]:
            fpath = os.path.join(run_folder, path_prefix, fname)
            if os.path.isfile(fpath):
                missing_columns = get_missing_columns(fpath, expected_columns)
        assert len(missing_columns) == 0, f"Missing columns in {fpath}: {missing_columns}"

    logger.debug("Validation completed!")


def get_missing_columns(fpath: str, column_names: list) -> list:
    """List of missing columns from a csv file.

    We just read the first row of a CSV to check the name of the columns

    Args:
        fpath: Path to the csv file
        column_names: list of columns to verify

    Returns
    -------
        A list of missing columns or empty list
    """
    try:
        _ = pd.read_csv(fpath, nrows=0).rename(columns=DEFAULT_COLUMN_MAP)
    except pd.errors.EmptyDataError:
        logger.error(f"Required file for R2X:{fpath} is empty!")
        raise

    return [col for col in column_names if col not in _.columns.str.lower()]


def get_missing_files(project_folder: str, file_list: Iterable, max_depth: int = 2) -> list:
    """List missing required files from project folder.

    This function looks recursively in the project folder. For safety we only
    look 2 levels of folders

    Args:
        project_folder: Folder to look for the files
        file_list: Iterable of files to check
        max_depth: Level of subfolders to look.

    Returns
    -------
        A list with the missing files or empty list
    """
    all_files = set()

    # Initialize stack with the project folder and depth 0
    input_folder = os.path.join(project_folder, "inputs_case")
    output_folder = os.path.join(project_folder, "outputs")
    stack: list[tuple[str, int]] = [(input_folder, 0), (output_folder, 0)]

    while stack:
        current_folder, current_depth = stack.pop()

        if current_depth > max_depth:
            continue

        for root, dirs, dir_files in os.walk(current_folder):
            for file_name in dir_files:
                file_path = os.path.join(root, file_name)
                all_files.add(os.path.basename(file_path))

            for folder in dirs:
                next_folder = os.path.join(root, folder)
                stack.append((next_folder, current_depth + 1))
    missing_files = [f for f in file_list if os.path.basename(f) not in all_files]
    return missing_files



"""R2X utils functions."""

# ruff: noqa
import io
import json
import ast
from operator import attrgetter

# Standard packages
import os
import shutil
from collections.abc import Iterable
from collections import ChainMap
import difflib
from importlib.resources import files
from pathlib import Path
from itertools import islice
from typing import Any, Hashable, Sequence

# Third-party packages
import numpy as np
import pandas as pd
import polars as pl
import pyarrow as pa
import pyarrow.dataset as ds
import pyarrow.parquet as pq
from tables import file
import yaml
from jsonschema import validate
from loguru import logger
import pint
from infrasys.base_quantity import BaseQuantity
from r2x.models import Generator


DEFAULT_OUTPUT_FOLDER: str = "r2x_export"
DEFAULT_DATA_FOLDER: str = "data"
DEFAULT_PLUGIN_PATH: str = "r2x.plugins"
DEFAULT_SEARCH_FOLDERS = [
    "outputs",
    "inputs_case",
    "outputs_perturb",
    "inputs_case/supplycurve_metadata",
    ".",
]


def get_project_root() -> Path:
    """Return the project absolute path.

    Args:
        as_str: Return the Path as string
    """
    return Path(__file__).parent.parent


def match_input_model(input_model: str) -> dict:
    match input_model:
        case "infrasys":
            fmap = {}
        case "reeds-US":
            fmap = read_fmap("r2x/defaults/reeds_us_mapping.json")
        case "reeds-India":
            fmap = read_fmap("r2x/defaults/india_mapping.json")
        case "sienna":
            fmap = read_fmap("r2x/defaults/sienna_mapping.json")
        case "plexos":
            fmap = read_fmap("r2x/defaults/plexos_mapping.json")
        case "nodal-plexos":
            fmap = (
                read_fmap("r2x/defaults/nodal_mapping.json")
                | read_fmap("r2x/defaults/plexos_mapping.json")
                | read_fmap("r2x/defaults/reeds_us_mapping.json")
            )
        case "nodal-sienna":
            fmap = (
                read_fmap("r2x/defaults/sienna_mapping.json")
                | read_fmap("r2x/defaults/nodal_mapping.json")
                | read_fmap("r2x/defaults/reeds_us_mapping.json")
            )
        case _:
            logger.error("Input model {} not recognized", input_model)
            raise KeyError(f"Input model {input_model=} not valid")
    return fmap


def validate_string(value):
    """Read cases flag value and convert it to Python type."""
    if value is None:
        return None
    try:
        return int(value)
    except ValueError:
        pass
    try:
        return float(value)
    except ValueError:
        pass
    if value == "true" or value == "TRUE":
        return True
    if value == "false" or value == "FALSE":
        return False
    # value = ast.literal_eval(value)
    # if len(value.split(",")) > 1:
    #     return value.split(",")
    try:
        value = ast.literal_eval(value)
    except:
        logger.trace("Could not parse {}", value)
    finally:
        return value


def get_max_value(
    df: pd.DataFrame,
    column: str,
    aggregator: list | str = "region",
) -> dict[str, float]:
    """Return max value of dataframe as a dict for a given aggregator.

    Args:
        df: Data to aggregate,
        column: Column to get,
        aggregator: Column to aggregate.
    """
    return df.max().to_dict()
    # return df.groupby(aggregator, observed=True)[column].max().to_dict()


def get_mean_data(
    data: pd.DataFrame,
    column: str,
    new_name: str,
    rename_dict: dict = {},
    categories=["tech", "tech_class", "region"],
) -> pd.DataFrame:
    """Return the mean value for a given dataframe and given categories.

    Args:
        data: Dataframe to aggregate,
        column: Column that will be renamed,
        new_name: New name of column,
        rename_dict: Dictionary passed to pd.DataFrame().rename(columns=rename_dict),
        categories: Keys to aggregate the data.

    Returns:
        Aggregated dataframe
    """
    if not rename_dict:
        rename_dict = {column: new_name}

    averaged_data = data.rename(columns=rename_dict).groupby(categories).agg(np.mean).reset_index()
    return averaged_data


def update_dict(base_dict: dict, override_dict: ChainMap | dict | None = None) -> dict:
    """Update or add defaults dictionary by overriding or creating new key."""
    if not override_dict:
        return base_dict
    _replace_keys = [
        "static_horizons",
        "static_models",
        "tech_map",
        "model_map",
        "plexos_fuel_map",
        "device_name_inference_map",
        "plexos_device_map",
    ]
    for key, value in override_dict.items():
        if key in base_dict and all(replace_key not in key for replace_key in _replace_keys):
            if isinstance(value, dict) and isinstance(base_dict[key], dict):
                update_dict(base_dict[key], value)  # Recursive call for nested dictionaries
            else:
                base_dict[key] = value  # Update the value or entire key-value pair
        elif key in base_dict:
            base_dict[key] = value
    return base_dict


def read_yaml(fname: str) -> dict:
    """Load Yaml to Python dictionary."""
    # Read configuration file
    with open(fname) as f:
        return yaml.safe_load(f)


def read_json(fname: str):
    """Load JSON file."""
    fpath = str(get_project_root())

    # Read configuration file
    with open(os.path.join(fpath, fname)) as f:
        return json.load(f)


def read_fmap(fname: str):
    """Read default fmap mapping for ReEDS files."""
    fmap = read_json(fname)
    validate(instance=fmap, schema=mapping_schema)

    # Lowercase dictionary
    fmap = {key.lower() if isinstance(key, str) else key: value for key, value in fmap.items()}
    return fmap


def get_missing_columns(fpath: str, column_names: list) -> list:
    """List of missing coluns from a csv file.

    We just read the first row of a CSV to check the name of the columns

    Args:
        fpath: Path to the csv file
        column_names: list of columns to verify

    Returns:
        A list of missing columns or empty list
    """
    try:
        _ = pd.read_csv(fpath, nrows=0).rename(columns=DEFAULT_COLUMN_MAP)
    except pd.errors.EmptyDataError:
        logger.error(f"Required file for R2X:{fpath} is empty!")
        raise

    return [col for col in column_names if col not in _.columns.str.lower()]


def get_missing_files(project_folder: str, file_list: Iterable, max_depth: int = 2) -> list:
    """List of missing required files from project folder.

    This function looks recursively in the project folder. For safety we only
    look 2 levels of folders

    Args:
        project_folder: Folder to look for the files
        file_list: Iterable of files to check
        max_depth: Level of subfolders to look.

    Returns:
        A list with the missing files or empty list
    """
    all_files = set()

    # Initialize stack with the project folder and depth 0
    stack: list[tuple[str, int]] = [(project_folder, 0)]

    while stack:
        current_folder, current_depth = stack.pop()

        if current_depth > max_depth:
            continue

        for root, dirs, dir_files in os.walk(current_folder):
            for file_name in dir_files:
                file_path = os.path.join(root, file_name)
                all_files.add(os.path.basename(file_path))

            for folder in dirs:
                next_folder = os.path.join(root, folder)
                stack.append((next_folder, current_depth + 1))
    missing_files = [f for f in file_list if os.path.basename(f) not in all_files]
    return missing_files


def read_csv(fname: str, package_data: str = "r2x.defaults", **kwargs) -> pl.LazyFrame:
    """Helper function to read csv string data from package data.

    Args:
        fname: Name of the csv file
        package_data: Location of file in package. Default location is r2x.defaults
        **kwargs: Additional keys passed to pandas read_csv function

    Returns:
        A pandas dataframe of the csv requested
    """
    csv_file = files(package_data).joinpath(fname).read_text(encoding="utf-8-sig")
    return pl.LazyFrame(pl.read_csv(io.StringIO(csv_file), **kwargs))


def get_timeindex(
    start_year: int = 2007, end_year: int = 2013, tz: str = "EST", year: int | None = None
) -> pd.DatetimeIndex:
    """ReEDS time indices are in EST, and leap years drop Dec 31 instead of Feb 29.

    Notes:
        - Function courtesy of P. Brown.

    Args:
        start_year: First year of data
        end_year: Last year of data
        tz: Time zone to convert to
        year: Create a 8760 profile for the given year
    Returns:
        Time index as numpy.ndarray
    """
    if year is not None:
        return pd.date_range(
            f"{year}-01-01",
            f"{year+1}-01-01",
            freq="H",
            inclusive="left",
            tz=tz,
        )[
            :8760  # This will remove extra day from leap years
        ]

    time_index = np.ravel(
        [
            pd.date_range(
                f"{y}-01-01",
                f"{y+1}-01-01",
                freq="H",
                inclusive="left",
                tz=tz,
            )[
                :8760  # This will remove extra day from leap years
            ]
            for y in range(start_year, end_year + 1)
        ]
    )
    return pd.to_datetime(time_index)


def clean_folder(path) -> None:
    """Create new folder if doesnt not exist. Purge if it does."""
    if os.path.exists(path):
        shutil.rmtree(path)
        os.makedirs(path)
    else:
        os.makedirs(path)
    return


def check_file_exists(
    fname: str,
    run_folder: str | os.PathLike,
    mandatory: bool = False,
) -> os.PathLike | None:
    """Return file path for a given filename if exists in folders.

    If it does not exists and is mandatory, it raises an FileNotFoundError

    Args:
        fname: Filename with extension of the desired file
        run_folder: Base folder of the run
        folders: Root folder name to start looking. Default: run_folder
        default_folders: Default location to look for files
        mandatory: Flag to identify needed files

    Returns:
        fpath or None
    """
    run_folder = Path(run_folder)
    # Set run_folder as default folder to look
    search_paths = [run_folder / folder for folder in DEFAULT_SEARCH_FOLDERS]

    file_matches = []
    for search_path in search_paths:
        if search_path.is_file():
            file_matches.append(search_path)
        for fpath in search_path.glob(fname):  # Rglob search recursively
            if fpath.name == fname:
                logger.trace("File '{}' found in {}", fname, fpath.parent)
                file_matches.append(fpath)

    if len(file_matches) > 1:
        msg = (
            f"Multiple files found for {fname}. Returning first match. "
            f"Check for copies of the files in the {file_matches}"
        )
        logger.warning(msg)

    if mandatory and not file_matches:
        raise FileNotFoundError(f"Mandatory file '{fname}' not found in {run_folder}.")
    elif not file_matches:
        logger.warning(f"File: '{fname}' not found in {run_folder}.")
        return None

    match = file_matches[0]
    return match


def get_csv(fpath: str, fname: str, fmap: dict[str, str | dict | list] = {}, **kwargs) -> None | pd.DataFrame:
    """Custom CSV file reader.

    This function reads a csv file using the fpath and save it into memory
    using the fname. In addition, it applies a custom column mapping if
    passed throught fmap dictionary.

    Args:
        fpath: Location of the file
        fname: Name or alias of the requested file
        fmap: File mapping values
        kwargs: Additional key arguments for pandas mostly

    Attributes:
        data:  ReEDS parsed data for PCM.
    """
    logger.debug(f"Attempting to read {fname}")

    config = read_json("r2x/defaults/config.json")
    column_mapping = config.get("default_column_mapping")
    custom_mapping = fmap.get("column_mapping", {})
    dtype = fmap.get("dtype", {})
    df_index = fmap.get("column_index", {})
    columns = fmap.get("columns", None)
    keep_case = fmap.get("keep_case", False)

    try:
        _ = pd.read_csv(fpath, header=0, **kwargs)
    except pd.errors.EmptyDataError:
        logger.warning(f"{fpath} is empty. Skipping it.")
        return None

    # This is a second safety mechanism if the file does infact has something
    # on it but no data.
    if _.empty:
        logger.warning(f"{fpath} is empty. Skipping it.")
        return None

    # Check if there are more columns added from the upgrader. For the moment,
    # we just append the s-region for validating the merge between different
    # dataframes.
    if columns is not None and isinstance(columns, list):
        # Assume that we only add one column for the moment.
        if len(_.columns) > len(columns):
            columns.append(_.columns[-1])

    # Reading CSV file
    data = (
        pd.read_csv(fpath, names=columns, header=0, **kwargs)  # type: ignore
        .rename(columns=str.lower)
        .rename(columns=column_mapping)
        .rename(columns=custom_mapping)
        .map(lambda r: r.lower() if isinstance(r, str) and not keep_case else r)
        .astype(dtype)
    )

    # Columns instead of rows. We dont know if in the future they are going to
    # change the format.  But just to be consistent with the other
    # variables that we read.
    match fname:
        case "years":
            data = data.T

    # Add index if passed.
    if df_index:
        try:
            data = data.set_index(df_index)
        except KeyError:
            raise KeyError(f"{df_index} not in {fname}.")

    # Create entry point to table
    logger.debug(f"Loaded file {fname}")
    return data


# NOTE: This is not actually used in the process but is kept here for documentation
# purposes.
def create_nodal_database(
    plexos_node_to_reeds_region_fpath: str,
    node_objects_fpath: str,
    reeds_shapefiles_fpath: str,
) -> pd.DataFrame:
    """Un-used function, but creates the nodes objects for the Nodal dataset."""
    import geopandas as gpd  # type: ignore

    nodes_to_ba = pd.read_csv(plexos_node_to_reeds_region_fpath)
    nodes_to_ba["node_id"] = nodes_to_ba["node"].str.split("_", n=1).str[0]
    nodes_to_ba = nodes_to_ba.loc[nodes_to_ba["node_id"].str.isnumeric()]
    nodes_to_ba = nodes_to_ba.groupby("node_id").agg({"reeds_region": list})
    nodes_to_ba["num_bas"] = nodes_to_ba["reeds_region"].str.len()
    nodes_to_ba = nodes_to_ba.loc[nodes_to_ba["num_bas"] == 1]
    nodes_to_ba["reeds_region"] = nodes_to_ba["reeds_region"].str[0]
    nodes_to_ba = nodes_to_ba.reset_index()[["node_id", "reeds_region"]]

    node_objects = node_objects_fpath
    usecols = ["Latitude", "Longitude", "Node ID", "kV", "Node", "Load Factor"]
    rename_dict = {
        "Node ID": "node_id",
        "kV": "voltage",
        "Load Factor": "load_participation_factor",
    }
    dtype = {
        "Node ID": "str",
        "Latitude": "float64",
        "Longitude": "float64",
        "kV": "float64",
        "Load Factor": "float64",
    }
    nodes = (
        pd.read_csv(node_objects, low_memory=False, dtype=dtype, usecols=usecols)  # type: ignore
        .rename(columns=rename_dict)
        .rename(columns=str.lower)
    )

    # Convert it to geopandas dataframe to assign the ReEDs region from the shapefile
    # NOTE: This step is not necessary if the node csv has the ReEDS region.
    reeds_shp = gpd.read_file(reeds_shapefiles_fpath)
    nodes = gpd.GeoDataFrame(nodes, geometry=gpd.points_from_xy(nodes.longitude, nodes.latitude)).set_crs(
        "epsg:4326"
    )
    nodes = gpd.sjoin(nodes, reeds_shp, predicate="within")

    nodes_to_bas = nodes.merge(nodes_to_ba, on="node_id", how="outer")
    nodes_to_bas.loc[nodes_to_bas["reeds_region"].isna(), "reeds_region"] = nodes_to_bas["pca"]
    nodes_to_bas.loc[
        (nodes_to_bas["reeds_region"] != nodes_to_bas["pca"]) & (~nodes_to_bas["pca"].isna()),
        "reeds_region",
    ] = nodes_to_bas["pca"]

    node_data = nodes_to_bas[
        [
            "node_id",
            "latitude",
            "longitude",
            "reeds_region",
            "voltage",
            "load_participation_factor",
        ]
    ]
    node_data["plexos_id"] = node_data["node_id"]
    return node_data


def invert_dict(d: dict[str, list]) -> dict[str, str]:
    """Inverse dictionary with list values.

    Example:
        dict = {"a": [1,2], "b":[3,4]}
        invers = {1:"a", 2: "a", 3:"b", 4 : "b"}
    """
    if d is None:
        return {}
    return {value: key for key, inner_list in d.items() for value in inner_list}


def match_category(row, categories, cutoff=0.6):
    """Return the n_return closest match based on the cutoff.

    This is a wrapper function of the difflib.get_close_matches function,
    except it just return the first value that if inds.

    Args:
        row: String to match,
        categories: categories to match against,
        n_return: number of matches to return,
        cutoff: Cutoff of the algorithm (0-1]. 1 being perfect match.
    """
    from difflib import get_close_matches

    result = get_close_matches(row, categories, n=1, cutoff=cutoff)

    if len(result) > 0:
        return result[0]
    return row


def get_defaults(
    input_model: str | None = None, output_model: str | None = None, *, verbose=None, **kwargs
) -> dict[str, str]:
    """Return configuration dictionary based on the output model."""
    config_dict = read_json("r2x/defaults/config.json")
    plugins_dict = read_json("r2x/defaults/plugins_config.json")

    config_dict = config_dict | plugins_dict

    if input_model is None and output_model is None:
        logger.debug("Returning base defaults")
        return config_dict

    # There is 4 paths for this to go:
    #   1. Zonal translations should always go throught "reeds-US" (as far as we only support one CEM"),
    #   2. If you want to translate a Plexos <-> Sienna model,
    #   3. If you want to read an existing infrasys system,
    #   4. If you want to run zonal to nodal.
    #       4.1 Plexos
    #       4.2 Sienna
    match input_model:
        case "infrasys":
            logger.debug("Returning infrasys defaults")
        case "reeds-US":
            config_dict = config_dict | read_json("r2x/defaults/reeds_input.json")
            logger.debug("Returning reeds defaults")
        case "reeds-India":
            config_dict = config_dict | read_json("r2x/defaults/nodal_defaults.json")
            logger.debug("Returning nodal defaults")
        case "nodal-sienna":
            config_dict = (
                config_dict
                | read_json("r2x/defaults/nodal_defaults.json")
                | read_json("r2x/defaults/pcm_defaults.json")
                | read_json("r2x/defaults/sienna_config.json")
                | read_json("r2x/defaults/reeds_input.json")
            )
            logger.debug("Returning nodal-sienna defaults")
        case "nodal-plexos":
            config_dict = (
                config_dict
                | read_json("r2x/defaults/nodal_defaults.json")
                | read_json("r2x/defaults/pcm_defaults.json")
                | read_json("r2x/defaults/plexos_input.json")
                | read_json("r2x/defaults/reeds_input.json")
            )
            logger.debug("Returning nodal-plexos defaults")
        case "sienna":
            config_dict = config_dict | read_json("r2x/defaults/sienna_config.json")
            logger.debug("Returning sienna defaults")
        case "plexos":
            config_dict = config_dict | read_json("r2x/defaults/plexos_input.json")
            logger.debug("Returning input_model {} defaults", input_model)
        case _:
            logger.warning("No input model passed")

    if output_model is None:
        return config_dict

    match output_model:
        case "plexos":
            pcm_dict = (
                read_json("r2x/defaults/plexos_output.json")
                | read_json("r2x/defaults/plexos_simulation_objects.json")
                | read_json("r2x/defaults/plexos_horizons.json")
                | read_json("r2x/defaults/plexos_models.json")
            )
            logger.debug("Returning output_model {} defaults", output_model)
        case "sienna":
            pcm_dict = read_json("r2x/defaults/sienna_config.json")
            logger.debug("Returning sienna defaults")
        case _:
            raise NotImplementedError(f"Model {output_model} not supported yet.")

    # Combine all in single dictionary. This assumes they do not share same key names.
    config_dict = config_dict | pcm_dict
    return config_dict


def get_enum_from_string(string: str, enum_class, prefix: str | None = None):
    max_similarity = 0.95
    closest_enum = None
    if prefix is None:
        prefix = ""
    for enum_member in enum_class:
        similarity = difflib.SequenceMatcher(None, prefix + string, enum_member.name).ratio()
        if similarity > max_similarity:
            max_similarity = similarity
            closest_enum = enum_member
    if closest_enum is None:
        raise KeyError(f"No matching enum found for string '{string}'")
    return closest_enum


def custom_attrgetter(component, category_attribute):
    try:
        category = attrgetter(category_attribute)(component)
    except AttributeError:
        category = category_attribute
    return category


def _unnest_all(schema, separator):
    def _unnest(schema, path=[]):
        for name, dtype in schema.items():
            base_type = dtype.base_type()

            if base_type == pl.Struct:
                yield from _unnest(dtype.to_schema(), [*path, name])
            else:
                yield [*path, name], dtype

    for (col, *fields), dtype in _unnest(schema):
        expr = pl.col(col)

        for field in fields:
            expr = expr.struct[field]

        if col == "":
            name = separator.join(fields)
        else:
            name = separator.join([col, *fields])

        yield expr.alias(name)


def unnest_all(df, separator="."):  # noqa: D103
    return df.select(_unnest_all(df.schema, separator))


def batched(lst, n):
    it = iter(lst)
    return iter(lambda: tuple(islice(it, n)), ())


def get_property_magnitude(property_value, to_unit: str | None = None) -> float:
    """Return magnitude with the given units for a pint Quantity.

    Parameters
    ----------
    property_name

    property_value
        pint.Quantity to extract magnitude from
    to_unit
        String that contains the unit conversion desired. Unit must be compatible.
    """
    if not isinstance(property_value, pint.Quantity | BaseQuantity):
        return property_value
    if to_unit:
        unit = to_unit.replace("$", "usd")  # Dollars are named usd on pint
        property_value = property_value.to(unit)
    return property_value.magnitude


DEFAULT_COLUMN_MAP = read_json("r2x/defaults/config.json").get("default_column_mapping")
mapping_schema = json.loads(files("r2x.defaults").joinpath("mapping_schema.json").read_text())



"""Electric load related models."""

from typing import Annotated

from pydantic import Field

from r2x.models.core import StaticInjection
from r2x.models.topology import ACBus, Bus
from r2x.units import ActivePower, ApparentPower


class ElectricLoad(StaticInjection):
    """Supertype for all electric loads."""

    bus: Bus = Field(description="Point of injection.")


class StaticLoad(ElectricLoad):
    """Supertype for static loads."""


class ControllableLoad(ElectricLoad):
    """Abstract class for controllable loads."""


class PowerLoad(StaticLoad):
    """Class representing a Load object."""

    active_power: (
        Annotated[
            ActivePower,
            Field(gt=0, description="Initial steady-state active power demand."),
        ]
        | None
    ) = None
    reactive_power: (
        Annotated[float, Field(gt=0, description="Reactive Power of Load at the bus in MW.")] | None
    ) = None
    max_active_power: Annotated[ActivePower, Field(gt=0, description="Max Load at the bus in MW.")] | None = (
        None
    )
    max_reactive_power: (
        Annotated[ActivePower, Field(gt=0, description=" Initial steady-state reactive power demand.")] | None
    ) = None
    base_power: Annotated[
        ApparentPower | None,
        Field(
            gt=0,
            description="Base power of the unit (MVA) for per unitization.",
        ),
    ] = None
    operation_cost: float | None = None

    @classmethod
    def example(cls) -> "PowerLoad":
        return PowerLoad(name="ExampleLoad", bus=ACBus.example(), active_power=ActivePower(1000, "MW"))


class InterruptiblePowerLoad(ControllableLoad):
    """A static interruptible power load."""

    base_power: Annotated[
        ApparentPower | None,
        Field(
            gt=0,
            description="Base power of the unit (MVA) for per unitization.",
        ),
    ] = None
    active_power: (
        Annotated[
            ActivePower,
            Field(gt=0, description="Initial steady-state active power demand."),
        ]
        | None
    ) = None
    reactive_power: (
        Annotated[float, Field(gt=0, description="Reactive Power of Load at the bus in MW.")] | None
    ) = None
    max_active_power: Annotated[ActivePower, Field(ge=0, description="Max Load at the bus in MW.")] | None = (
        None
    )
    max_reactive_power: (
        Annotated[ActivePower, Field(gt=0, description=" Initial steady-state reactive power demand.")] | None
    ) = None
    operation_cost: float | None = None



"""Script that creates simple systems for testing."""

from datetime import datetime, timedelta

from infrasys.cost_curves import FuelCurve
from infrasys.function_data import LinearFunctionData
from infrasys.time_series_models import SingleTimeSeries
from infrasys.value_curves import InputOutputCurve
from r2x.api import System
from r2x.enums import PrimeMoversType
from r2x.models import (
    ACBus,
    Area,
    GenericBattery,
    LoadZone,
    MonitoredLine,
    RenewableDispatch,
    ThermalStandard,
)
from r2x.models.costs import ThermalGenerationCost
from r2x.units import Energy, Percentage, Time, ureg


def ieee5bus() -> System:
    """Return an instance of the IEE 5-bus system."""
    system = System(name="IEEE 5-bus System", auto_add_composed_components=True)

    area_1 = Area(name="region1")
    load_zone_1 = LoadZone(name="LoadZone1")
    bus_1 = ACBus(number=1, name="node_a", base_voltage=100 * ureg.volt, area=area_1, load_zone=load_zone_1)
    bus_2 = ACBus(number=2, name="node_b", base_voltage=100 * ureg.volt, area=area_1, load_zone=load_zone_1)
    bus_3 = ACBus(number=3, name="node_c", base_voltage=100 * ureg.volt, area=area_1, load_zone=load_zone_1)
    bus_4 = ACBus(number=4, name="node_d", base_voltage=100 * ureg.volt, area=area_1, load_zone=load_zone_1)
    bus_5 = ACBus(number=5, name="node_e", base_voltage=100 * ureg.volt, area=area_1, load_zone=load_zone_1)
    # Append buses generator
    for bus in [bus_1, bus_2, bus_3, bus_4, bus_5]:
        system.add_component(bus)

    # Solar generators
    initial_time = datetime(year=2012, month=1, day=1)
    ts = SingleTimeSeries.from_array(
        data=range(0, 8760),
        initial_time=initial_time,
        resolution=timedelta(hours=1),
        variable_name="rated_capacity",
    )
    solar_pv_01 = RenewableDispatch(
        name="SolarPV1",
        bus=bus_3,
        prime_mover_type=PrimeMoversType.PV,
        active_power=384 * ureg.MW,
        category="solar",
    )
    solar_pv_02 = RenewableDispatch(
        name="SolarPV2",
        bus=bus_4,
        prime_mover_type=PrimeMoversType.PV,
        active_power=384 * ureg.MW,
        category="solar",
    )
    system.add_component(solar_pv_01)
    system.add_component(solar_pv_02)
    system.add_time_series(ts, solar_pv_01, solar_pv_02)

    # Storage
    storage = GenericBattery(
        name="Battery1",
        bus=bus_2,
        prime_mover_type=PrimeMoversType.BA,
        active_power=200 * ureg.MW,
        charge_efficiency=Percentage(85, "%"),
        storage_capacity=Energy(800, "MWh"),
        storage_duration=Time(4, "h"),
        category="storage",
    )
    system.add_component(storage)

    # Thermal generators
    alta = ThermalStandard(
        name="Alta",
        fuel="gas",
        prime_mover_type=PrimeMoversType.CC,
        active_power=40 * ureg.MW,
        min_rated_capacity=10 * ureg.MW,
        bus=bus_1,
        category="thermal",
    )
    system.add_component(alta)
    brighton = ThermalStandard(
        name="Brighton",
        fuel="Gas",
        prime_mover_type=PrimeMoversType.CC,
        active_power=600 * ureg.MW,
        min_rated_capacity=150 * ureg.MW,
        bus=bus_5,
        category="thermal",
    )
    system.add_component(brighton)
    park_city = ThermalStandard(
        name="Park City",
        fuel="Gas",
        prime_mover_type=PrimeMoversType.CC,
        active_power=170 * ureg.MW,
        min_rated_capacity=20 * ureg.MW,
        operation_cost=ThermalGenerationCost(
            variable=FuelCurve(
                value_curve=InputOutputCurve(
                    function_data=LinearFunctionData(proportional_term=10, constant_term=0)
                ),
                fuel_cost=15,
            )
        ),
        bus=bus_1,
        category="thermal",
    )
    system.add_component(park_city)
    solitude = ThermalStandard(
        name="Solitude",
        fuel="Gas",
        prime_mover_type=PrimeMoversType.CC,
        active_power=520 * ureg.MW,
        operation_cost=ThermalGenerationCost(
            variable=FuelCurve(
                value_curve=InputOutputCurve(
                    function_data=LinearFunctionData(proportional_term=10, constant_term=0)
                ),
                fuel_cost=15,
            )
        ),
        bus=bus_3,
        category="thermal",
    )
    system.add_component(solitude)
    sundance = ThermalStandard(
        name="Sundance",
        fuel="Gas",
        prime_mover_type=PrimeMoversType.CC,
        active_power=400 * ureg.MW,
        min_rated_capacity=80 * ureg.MW,
        bus=bus_4,
        category="thermal",
    )
    system.add_component(sundance)

    # Branch
    branch_ab = MonitoredLine(name="line_ab", from_bus=bus_1, to_bus=bus_2, rating_up=400 * ureg.MW)
    system.add_component(branch_ab)

    branch_ad = MonitoredLine(name="line_ad", from_bus=bus_1, to_bus=bus_4, rating_up=400 * ureg.MW)
    system.add_component(branch_ad)

    branch_ae = MonitoredLine(name="line_ae", from_bus=bus_1, to_bus=bus_5, rating_up=400 * ureg.MW)
    system.add_component(branch_ae)

    branch_bc = MonitoredLine(name="line_bc", from_bus=bus_2, to_bus=bus_3, rating_up=400 * ureg.MW)
    system.add_component(branch_bc)

    branch_cd = MonitoredLine(name="line_cd", from_bus=bus_3, to_bus=bus_4, rating_up=400 * ureg.MW)
    system.add_component(branch_cd)

    branch_ed = MonitoredLine(name="line_ed", from_bus=bus_5, to_bus=bus_4, rating_up=240 * ureg.MW)
    system.add_component(branch_ed)

    return system



"""CLI helper functions."""

import argparse
import importlib
from importlib.resources import files
from .__version__ import __version__


class Flags(argparse.Action):
    """Class to enable feature flags on the code and CLI.

    This will save the feature flags in a `flags` argument in the NameSpace of
    argparse
    """

    def __call__(  # type: ignore
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: str,
        optional_str: str,
    ) -> None:
        """Save flags as arguments in Namespace."""
        setattr(namespace, self.dest, dict())

        for value in values:
            # split it into key and value
            key, value = value.split("=")
            # assign into dictionary
            getattr(namespace, self.dest)[key] = value


def get_additional_arguments(
    parser: argparse.ArgumentParser,
    folder: str = "r2x.plugins",
) -> argparse.ArgumentParser:
    """Add plugin arguments to CLI.

    Args:
        parser: CLI parser object.
        plugin_list: List of plugins.
        folder: Folder that contains the plugin.
    """
    plugin_files = [plugin for plugin in files(folder).rglob("*.py") if "__" not in plugin.name]  # type: ignore
    for plugin in plugin_files:
        plugin_module = importlib.import_module(f".{plugin.stem}", folder)
        if hasattr(plugin_module, "cli_arguments"):
            plugin_cli_group = parser.add_argument_group(f"Plugin {plugin.stem}")
            plugin_module.cli_arguments(plugin_cli_group)

    parser_files = [parser for parser in files("r2x.parser").rglob("*.py")]  # type: ignore
    for parser_file in parser_files:
        parser_module = importlib.import_module(f".{parser_file.stem}", "r2x.parser")
        if hasattr(parser_module, "cli_arguments"):
            parser_cli_group = parser.add_argument_group(f"Parser {parser_file.stem}")
            parser_module.cli_arguments(parser_cli_group)
    return parser


def base_cli() -> argparse.ArgumentParser:
    """Create parser object for CLI."""
    parser = argparse.ArgumentParser(
        description="""Execute translation of ReEDS scenario results into X model""",
        add_help=True,
        prog="r2x",
    )

    group_run = parser.add_argument_group("Options for running the code")
    group = group_run.add_mutually_exclusive_group()
    group.add_argument(
        "-i",
        dest="run_folder",
        help="Location of the folder with the input model data",
    )
    group.add_argument(
        "--cases",
        dest="cases_file",
        help="Path to the cases file",
    )
    group_cli = parser.add_argument_group("Args used alongside run folder")
    group_cli.add_argument(
        "--name",
        dest="name",
        help="Scenario name",
    )
    group_cli.add_argument(
        "--input-model",
        dest="input_model",
        choices=[
            "plexos",
            "reeds-US",
            "plexos",
            "infrasys",
        ],
        help="Input model to convert from",
    )
    group_cli.add_argument(
        "--output-model",
        dest="output_model",
        choices=["plexos", "sienna", "pras"],
        help="Output model to convert to",
    )
    group_cli.add_argument(
        "-o",
        nargs="?",
        dest="output_folder",
        help="Path to save the result translation",
    )
    group_cli.add_argument(
        "-y",
        "--year",
        type=int,
        nargs="+",
        dest="solve_year",
        help="Year to translate",
    )
    group_cli.add_argument(
        "--weather-year",
        type=int,
        dest="weather_year",
        help="Year to translate",
    )
    group_cli.add_argument(
        "--scenario",
        dest="scenario_name",
        help="Scenario to select from cases file",
    )
    group_cli.add_argument(
        "-ud",
        "--user-dict",
        dest="user_dict",
        help="User provided overrides to defaults",
    )
    group_cli.add_argument(
        "-p",
        "--plugins",
        dest="plugins",
        nargs="*",
        help="Plugins to be included in the translation",
    )
    group_cli.add_argument(
        "--save",
        action="store_true",
        help="Store system in the same output folder",
    )
    group_cli.add_argument("--inspect", action="store_true", help="Inspect infrasys system")
    group_cli.add_argument("--upgrade", action="store_true", help="Run upgrader")
    group_cli.add_argument("--flags", nargs="*", dest="feature_flags", action=Flags, help="Feature flags")
    parser.add_argument("--debug", action="store_true", help="Enable debug logging")
    parser.add_argument("--version", action="version", version=f"R2X version: {__version__}")
    parser = get_additional_arguments(parser)
    return parser



# Models
# ruff: noqa
from .branch import Branch, ACBranch, DCBranch, MonitoredLine, Transformer2W
from .core import ReserveMap, TransmissionInterfaceMap
from .generators import (
    Generator,
    ThermalGen,
    Storage,
    RenewableGen,
    GenericBattery,
    HybridSystem,
    HydroGen,
    HydroPumpedStorage,
    HydroDispatch,
    HydroEnergyReservoir,
    RenewableDispatch,
    RenewableNonDispatch,
    ThermalStandard,
    Storage,
)
from .costs import HydroGenerationCost, StorageCost, ThermalGenerationCost, RenewableGenerationCost
from .load import PowerLoad, InterruptiblePowerLoad
from .services import Emission, Reserve, TransmissionInterface
from .topology import ACBus, Area, Bus, LoadZone
